import pytest

from expanse.common.configuration.config import Config
from expanse.core.application import Application
from expanse.core.bootstrap.load_configuration import LoadConfiguration


@pytest.fixture()
def app() -> Application:
    return Application()


def test_loading_configuration_provides_default_values(app: Application) -> None:
    LoadConfiguration.bootstrap(app)

    config = app.container.get(Config)

    assert config["app.name"] == "Expanse"
    assert len(config["app.providers"]) > 0



from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from expanse.core.application import Application


class RegisterProviders:
    @classmethod
    def bootstrap(cls, app: Application) -> None:
        app.register_configured_providers()



from collections.abc import Awaitable
from collections.abc import Callable
from collections.abc import MutableMapping
from typing import Any


Scope = MutableMapping[str, Any]
Message = MutableMapping[str, Any]

Receive = Callable[[], Awaitable[Message]]
Send = Callable[[Message], Awaitable[None]]

ASGIApp = Callable[[Scope, Receive, Send], Awaitable[None]]



from __future__ import annotations

import re

from importlib import import_module
from importlib.util import module_from_spec
from importlib.util import spec_from_file_location
from types import NoneType
from typing import TYPE_CHECKING
from typing import Any
from typing import ForwardRef
from typing import _eval_type  # type: ignore[attr-defined]


if TYPE_CHECKING:
    from pathlib import Path
    from types import ModuleType


def string_to_class(string: str) -> type[Any]:
    method_name: str | None = None
    parts = string.split(":", maxsplit=1)
    if len(parts) > 1:
        method_name = parts[-1]

    module_name, class_name = parts[0].rsplit(".", maxsplit=1)

    module = import_module(module_name)

    class_ = getattr(module, class_name)

    if not method_name:
        return class_

    return getattr(class_, method_name)


def module_from_path(path: Path, name: str | None = None) -> ModuleType | None:
    spec = spec_from_file_location(name or path.with_suffix("").name, path)

    if spec is None:
        return None

    module = module_from_spec(spec)

    if module is None or spec.loader is None:
        return None

    spec.loader.exec_module(module)

    return module


def string_matches(string: str, pattern: str | list[str]) -> bool:
    if isinstance(pattern, str):
        pattern = [pattern]

    for pat in pattern:
        if string == pat:
            return True

        pat = re.escape(pat)
        pat = pat.replace(r"\*", ".*")

        if re.match(f"^{pat}", string):
            return True

    return False


def class_to_name(class_: type | str) -> str:
    if isinstance(class_, str):
        return class_

    module = class_.__module__
    name = class_.__qualname__

    full_name = f"{module}.{name}" if module else name

    return full_name


def eval_type_lenient(
    value: Any, globalns: dict[str, Any] | None, localns: dict[str, Any] | None
) -> Any:
    if value is None:
        value = NoneType
    elif isinstance(value, str):
        value = ForwardRef(value, is_argument=False, is_class=True)

    try:
        return _eval_type(value, globalns, localns)
    except NameError:
        # the point of this function is to be tolerant to this case
        return value



from collections.abc import MutableMapping
from os import PathLike
from typing import Any
from typing import NoReturn

from baize.asgi import FileResponse
from baize.asgi import JSONResponse
from baize.asgi import PlainTextResponse

from expanse.asynchronous.http.response import Response
from expanse.asynchronous.routing.redirect import Redirect
from expanse.asynchronous.view.view_factory import ViewFactory
from expanse.common.core.http.exceptions import HTTPException


class Responder:
    def __init__(self, view: ViewFactory, redirect: Redirect) -> None:
        self._view = view
        self._redirect = redirect

    def redirect(self) -> Redirect:
        return self._redirect

    async def text(
        self,
        content: str = "",
        status_code: int = 200,
        headers: MutableMapping[str, Any] | None = None,
        media_type: str = "text/plain",
    ) -> Response:
        return Response(
            response=PlainTextResponse(
                content, status_code=status_code, headers=headers, media_type=media_type
            )
        )

    async def html(
        self,
        content: str = "",
        status_code: int = 200,
        headers: MutableMapping[str, Any] | None = None,
    ) -> Response:
        return await self.text(
            content, status_code=status_code, headers=headers, media_type="text/html"
        )

    async def json(
        self,
        content: Any = "",
        status_code: int = 200,
        headers: MutableMapping[str, Any] | None = None,
        **kwargs: Any,
    ) -> Response:
        return Response(
            response=JSONResponse(
                content, status_code=status_code, headers=headers, **kwargs
            )
        )

    async def file(
        self,
        path: str | PathLike[str],
        headers: MutableMapping[str, str] | None = None,
        media_type: str | None = None,
        filename: str | None = None,
        chunk_size: int = 4096 * 64,
    ) -> Response:
        return Response(
            response=FileResponse(
                str(path),
                headers=headers,
                content_type=media_type,
                download_name=filename,
                chunk_size=chunk_size,
            )
        )

    async def view(
        self,
        view: str,
        data: MutableMapping[str, Any] | None = None,
        status_code: int = 200,
        headers: MutableMapping[str, Any] | None = None,
    ) -> Response:
        return await self._view.render(
            await self._view.make(view, data, status_code, headers)
        )

    async def abort(
        self,
        status_code: int,
        message: str | None = None,
        headers: MutableMapping[str, Any] | None = None,
    ) -> NoReturn:
        raise HTTPException(status_code, detail=message, headers=headers)



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Protocol


if TYPE_CHECKING:
    from expanse.asynchronous.core.application import Application


class Bootstrapper(Protocol):
    @classmethod
    async def bootstrap(cls, app: Application) -> None: ...



from typing import Annotated

from sqlalchemy import BigInteger
from sqlalchemy import Identity
from sqlalchemy.dialects import sqlite
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import MappedAsDataclass

from expanse.common.database.orm import column


primary_key = Annotated[
    int,
    column(
        BigInteger().with_variant(sqlite.INTEGER(), "sqlite"),
        Identity(always=True),
        primary_key=True,
    ),
]


class Model(MappedAsDataclass, DeclarativeBase): ...


class User(Model):
    __tablename__: str = "users"

    id: Mapped[primary_key] = column()
    first_name: Mapped[str] = column()
    last_name: Mapped[str | None] = column(default=None)
    email: Mapped[str] = column()



from __future__ import annotations

from typing import TYPE_CHECKING

from expanse.routing.redirect import Redirect
from expanse.routing.router import Router
from expanse.support.service_provider import ServiceProvider


if TYPE_CHECKING:
    from expanse.common.http.url_path import URLPath
    from expanse.view.view_factory import ViewFactory


class RoutingServiceProvider(ServiceProvider):
    def register(self) -> None:
        self._container.singleton(Router, lambda container: Router(container))
        self._container.alias(Router, "router")
        self._container.scoped(Redirect)

    def boot(self) -> None:
        self._container.on_resolved("view", self._register_view_globals)

    def _register_view_globals(self, view: ViewFactory) -> None:
        router = self._container.get(Router)

        def route(name: str, **parameters) -> URLPath:
            return router.route(name, parameters)

        def url(name: str, **parameters) -> URLPath:
            return router.url(name, parameters)

        view.register_global(url=url)
        view.register_global(route=route)



from typing import TYPE_CHECKING

from expanse.asynchronous.static.static import Static
from expanse.asynchronous.support.service_provider import ServiceProvider
from expanse.asynchronous.view.view_factory import ViewFactory
from expanse.common.configuration.config import Config


if TYPE_CHECKING:
    from pathlib import Path

    from expanse.asynchronous.core.application import Application
    from expanse.asynchronous.routing.router import Router


class StaticServiceProvider(ServiceProvider):
    async def register(self) -> None:
        self._container.singleton(Static, self._register_static)

    async def boot(self) -> None:
        await self._container.on_resolved("router", self._add_static_route)
        await self._container.on_resolved("view", self._register_view_globals)

    async def _register_static(self, config: Config) -> Static:
        app: Application = await self._container.get("app")
        paths: list[Path] = config.get("static.paths", [])
        paths = [app.base_path / p if not p.is_absolute() else p for p in paths]
        if (url := config.get("static.url")) is not None:
            url = str(url)

        return Static(paths, prefix=config.get("static.prefix"), url=url)

    async def _add_static_route(self, router: "Router") -> None:
        config = await self._container.get(Config)
        if config.get("app.debug", False):
            prefix: str = config["static.prefix"].rstrip("/")

            router.get(
                f"{prefix}/{{path:path}}",
                (await self._container.get(Static)).get,
                name="static",
            )

    async def _register_view_globals(self, view: ViewFactory) -> None:
        view.register_global(static=(await self._container.get(Static)).url)



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Annotated
from typing import Self
from typing import TypeVar

from pydantic import BaseModel


T = TypeVar("T", bound=BaseModel)

if TYPE_CHECKING:
    from typing import Union

    Query = Union[T, T]  # noqa: UP007
else:

    class Query:
        def __class_getitem__(cls, item: type[T]) -> Annotated[type[T], type[Self]]:
            return Annotated[item, cls]



from typing import Any

from sqlalchemy import URL
from sqlalchemy import event
from sqlalchemy import make_url
from sqlalchemy.ext.asyncio import async_sessionmaker
from sqlalchemy.util import immutabledict

from expanse.asynchronous.core.application import Application
from expanse.asynchronous.database.connection import Connection
from expanse.asynchronous.database.engine import Engine
from expanse.asynchronous.database.session import Session
from expanse.common.database._utils import create_engine
from expanse.database.config import DatabaseConfig
from expanse.database.config import PostgreSQLConfig
from expanse.database.config import SQLiteConfig


class DatabaseManager:
    def __init__(self, app: Application) -> None:
        self._app: Application = app
        self._engines: dict[str, Engine] = {}
        self._factories: dict[str, async_sessionmaker] = {}

    def connection(self, name: str | None = None) -> Connection:
        engine = self.configure_engine(name)

        connection = engine.connect()

        assert isinstance(connection, Connection)

        return connection

    def session(self, name: str | None = None) -> Session:
        name = name or self.get_default_connection()

        if name in self._factories:
            return self._factories[name]()

        engine = self.configure_engine(name)
        factory = async_sessionmaker(engine, class_=Session)

        self._factories[name] = factory

        return self._factories[name]()

    def create_base_engine(self, url: URL, **kwargs) -> Engine:
        sync_engine = create_engine(url, **kwargs)

        return Engine(sync_engine)

    def configure_engine(self, name: str | None = None) -> Engine:
        name = name or self.get_default_connection()

        if name in self._engines:
            return self._engines[name]

        config = self._configuration(name)

        if not config:
            raise

        self._engines[name] = self._create_engine(config)

        return self._engines[name]

    def _create_engine(self, raw_config: dict[str, Any]) -> Engine:
        config = DatabaseConfig.model_validate(raw_config).root

        match config:
            case SQLiteConfig():
                return self._create_sqlite_engine(config)

            case PostgreSQLConfig():
                return self._create_postgresql_engine(config)

    def _create_sqlite_engine(self, config: SQLiteConfig) -> Engine:
        if config.url is not None:
            url = make_url(str(config.url))
            if url.drivername == "sqlite":
                url = URL("sqlite+aiosqlite", *url[1:])
        else:
            if config.database is None:
                raise ValueError("The SQLite database path is not configured.")

            database_path = config.database

            database: str
            if database_path == ":memory:":
                database = database_path
            else:
                if not database_path.is_absolute():
                    database_path = self._app.base_path / database_path

                database_path.parent.mkdir(parents=True, exist_ok=True)

                database = database_path.as_posix()

            url = URL(
                drivername="sqlite+aiosqlite",
                database=database,
                host=None,
                port=None,
                username=None,
                password=None,
                query=immutabledict(),
            )

        engine = self.create_base_engine(url)

        if config.foreign_key_constraints:

            @event.listens_for(engine.sync_engine, "connect")
            def set_sqlite_pragma(dbapi_connection, connection_record):
                cursor = dbapi_connection.cursor()
                cursor.execute("PRAGMA foreign_keys=ON")
                cursor.close()

        return engine

    def _create_postgresql_engine(self, config: PostgreSQLConfig) -> Engine:
        if config.url is not None:
            url = make_url(str(config.url))
            if url.drivername in {"postgresql", "postgresql+psycopg"}:
                url = URL("postgresql+psycopg_async", *url[1:])
        else:
            drivername: str = "postgresql"
            if config.dbapi is not None:
                dbapi = config.dbapi
                if dbapi == "psycopg":
                    dbapi = "psycopg_async"

                drivername += f"+{dbapi}"
            else:
                drivername += "+psycopg_async"

            query: dict[str, Any] = {}

            if config.sslmode is not None:
                query["sslmode"] = config.sslmode

            url = URL(
                drivername=drivername,
                host=config.host,
                port=config.port,
                database=config.database,
                username=config.username,
                password=config.password,
                query=immutabledict(query),
            )

        engine = self.create_base_engine(
            url, **config.pool.model_dump(exclude_none=True)
        )

        if config.search_path:

            @event.listens_for(engine.sync_engine, "connect", insert=True)
            def set_search_path(dbapi_connection, connection_record):
                existing_autocommit = dbapi_connection.autocommit
                dbapi_connection.autocommit = True
                cursor = dbapi_connection.cursor()
                cursor.execute(f"SET SESSION search_path='{config.search_path}'")
                cursor.close()
                dbapi_connection.autocommit = existing_autocommit

        return engine

    def get_default_connection(self) -> str:
        return self._app.config.get("database.default")

    def _configuration(self, name: str) -> Any:
        connections = self._app.config.get("database.connections", {})

        if name not in connections:
            raise ValueError(f"The database connection [{name}] not configured.")

        return connections[name]



from typing import Any

from expanse.common.configuration.config import Config
from expanse.http.response import Response
from expanse.http.response_adapter import ResponseAdapter
from expanse.support.service_provider import ServiceProvider
from expanse.view.view import View
from expanse.view.view_factory import ViewFactory
from expanse.view.view_finder import ViewFinder


class ViewServiceProvider(ServiceProvider):
    def register(self) -> None:
        self._register_factory()
        self._register_view_finder()

    def boot(self) -> None:
        self._container.on_resolved(ResponseAdapter, self._register_response_adapters)

    def _register_factory(self) -> None:
        self._container.singleton(ViewFactory, self._create_factory)
        self._container.alias(ViewFactory, "view")

    def _create_factory(self, finder: ViewFinder, config: Config) -> ViewFactory:
        return ViewFactory(finder, debug=config.get("app.debug", False))

    def _register_view_finder(self) -> None:
        def _create_view_finder(config: Config) -> ViewFinder:
            return ViewFinder(config["view"]["paths"])

        self._container.singleton(ViewFinder, _create_view_finder)
        self._container.alias(ViewFinder, "view:finder")

    def _register_response_adapters(self, adapter: ResponseAdapter) -> None:
        def adapt_view(
            raw_response: View,
            factory: ViewFactory,
            declared_response_type: type[Any] | None = None,
        ) -> Response:
            return factory.render(raw_response)

        adapter.register_adapter(View, adapt_view)



from pathlib import Path
from typing import Annotated
from typing import Literal

from pydantic import AnyUrl
from pydantic import BaseModel
from pydantic import Field
from pydantic import RootModel
from pydantic_settings import SettingsConfigDict


class PoolConfig(BaseModel):
    pool_pre_ping: bool | None = None
    pool_size: int | None = None
    pool_timeout: float | None = None
    pool_recycle: int | None = None
    max_overflow: int | None = None


class SQLiteConfig(PoolConfig, BaseModel):
    driver: Literal["sqlite"] = "sqlite"
    url: AnyUrl | None = None
    database: Path | Literal[":memory:"] | None = None
    foreign_key_constraints: bool = Field(default=True, alias="foreign_keys")

    model_config = SettingsConfigDict(arbitrary_types_allowed=True)


class PostgreSQLConfig(PoolConfig, BaseModel):
    driver: Literal["postgresql"] = "postgresql"
    dbapi: (
        Literal["psycopg", "psycopg2", "pg8000", "psycopg_async", "asyncpg"] | None
    ) = None
    url: AnyUrl | None = None
    host: str | None = None
    port: int = 5432
    database: str | None = None
    username: str = ""
    password: str = ""
    search_path: str | None = None
    sslmode: Literal["prefer", "require"] | None = None
    pool: PoolConfig = PoolConfig()


class DatabaseConfig(RootModel):
    root: Annotated[SQLiteConfig | PostgreSQLConfig, Field(discriminator="driver")]



import re

from argparse import Namespace
from collections.abc import Iterator
from contextlib import contextmanager
from contextlib import suppress
from pathlib import Path

from alembic import util
from alembic.command import downgrade
from alembic.command import init
from alembic.command import revision
from alembic.command import upgrade
from cleo.io.io import IO
from cleo.io.null_io import NullIO
from cleo.ui.progress_indicator import ProgressIndicator

from expanse.common.core.application import Application
from expanse.common.database.migration.config import Config as AlembicConfig
from expanse.common.database.migration.utils import configure_alembic_loggers


class Buffer:
    def __init__(self, io: IO) -> None:
        from pygments.formatters import TerminalFormatter
        from pygments.lexers import get_lexer_by_name

        self._io = io
        self._output = io.output
        self._lexer = get_lexer_by_name("sql", stripnl=False)
        self._formatter = TerminalFormatter()

    def write(self, data: str) -> int:
        if self._should_filter_message(data):
            return 0

        data = self._format(data)
        self._io.write(data)

        return len(data)

    def read(self, size: int = -1) -> str:
        return self._io.read(size)

    def flush(self) -> None:
        self._io.flush()

    def isatty(self) -> bool:
        return False

    def seekable(self) -> bool:
        return False

    def readable(self) -> bool:
        return True

    def writable(self) -> bool:
        return True

    def close(self) -> None:
        self.flush()

    def _format(self, data: str) -> str:
        if not self._io.is_decorated():
            return data

        from pygments import highlight

        return highlight(data, self._lexer, self._formatter)

    def _should_filter_message(self, message: str) -> bool:
        if message == "Generating static SQL":
            return True

        return False


class Migrator:
    def __init__(self, app: Application) -> None:
        self._app: Application = app
        self._config = AlembicConfig(app, cmd_opts=Namespace(quiet=True))

    @property
    def config(self) -> AlembicConfig:
        return self._config

    def init(self, io: IO | None = None) -> None:
        if io is None:
            io = NullIO()

        path = self._app.named_path("database").joinpath("migrations")
        with suppress(ValueError):
            path = path.relative_to(self._app.base_path)

        with self._patched_status(io):
            init(
                self._config,
                str(path),
                template="default",
            )

    def make(
        self, message: str | None = None, auto: bool = False, *, io: IO | None = None
    ) -> None:
        if io is None:
            io = NullIO()

        migration_directory = self._app.named_path("database").joinpath("migrations")

        if not migration_directory.exists():
            self.init(io=io)

        if auto:
            # If the migration should be auto-generated, load the models first
            # to ensure that alembic sees all the model changes.
            # This is not done in the Alembic's env.py file to avoid loading
            # models unnecessarily.
            self._load_models()

        with self._patched_status(io):
            revision(self._config, message=message, autogenerate=auto)

    def migrate(
        self, revision: str = "head", dry_run: bool = False, *, io: IO | None = None
    ) -> None:
        if io is None:
            io = NullIO()

        if dry_run:
            configure_alembic_loggers(io, disable=True)
            self._config.output_buffer = Buffer(io)  # type: ignore[assignment]

        upgrade(self._config, revision, sql=dry_run)

    def rollback(self, revision: str = "-1", *, io: IO | None = None) -> None:
        if io is None:
            io = NullIO()

        downgrade(self._config, revision)

    def _load_models(self) -> None:
        model_directory = self._app.path("models")
        for path in model_directory.rglob("*.py"):
            if path.stem.startswith("_"):
                continue

            from importlib import import_module

            path = path.relative_to(self._app.base_path)
            module_name = path.with_suffix("").as_posix().replace("/", ".")
            import_module(module_name)

    @contextmanager
    def _indicator(
        self,
        io: IO,
        start_message: str,
        end_message: str,
        fmt: str = "{message}",
    ) -> Iterator[None]:
        indicator = ProgressIndicator(io, fmt)

        with indicator.auto(start_message, end_message):
            yield

    @contextmanager
    def _patched_status(self, io: IO) -> Iterator[None]:
        @contextmanager
        def status(
            status_msg: str, newline: bool = False, quiet: bool = False
        ) -> Iterator[None]:
            if status_msg.startswith("Generating "):
                try:
                    path = Path(status_msg[11:]).relative_to(self._app.base_path)
                    status_msg = f"Generating <c1>{path}</c1>"
                except Exception:
                    pass
            elif status_msg.startswith("Creating directory "):
                try:
                    if m := re.match(r"Creating directory '([^']+)'", status_msg):
                        try:
                            path = Path(m.group(1)).relative_to(self._app.base_path)
                        except ValueError:
                            # The path is not relative to the base path
                            # Most likely due to being run in tests
                            path = Path(m.group(1))

                        status_msg = f"Creating directory <c1>{path}</c1>"
                except Exception:
                    pass

            status_msg = f"{status_msg}..."

            with self._indicator(
                io,
                status_msg,
                status_msg + " <success>Done</success>",
                fmt="  - {message}",
            ):
                yield

        original_status = util.status
        util.status = status

        yield

        util.status = original_status



import re
import sys

from collections.abc import Callable
from pathlib import Path
from typing import ClassVar

from cleo.io.inputs.input import Input
from cleo.io.inputs.string_input import StringInput
from cleo.io.io import IO
from cleo.io.outputs.buffered_output import BufferedOutput
from cleo.io.outputs.output import Output
from cleo.io.outputs.stream_output import StreamOutput

import expanse

from expanse.common.support._utils import string_to_class
from expanse.console.commands.command import Command
from expanse.console.console import Console
from expanse.contracts.debug.exception_handler import ExceptionHandler
from expanse.core.application import Application
from expanse.core.bootstrap.boot_providers import BootProviders
from expanse.core.bootstrap.bootstrapper import Bootstrapper
from expanse.core.bootstrap.load_configuration import LoadConfiguration
from expanse.core.bootstrap.load_environment_variables import LoadEnvironmentVariables
from expanse.core.bootstrap.register_providers import RegisterProviders


class Gateway:
    """
    The gateway is the layer between the command line and Expanse internal architecture.
    """

    _bootstrappers: ClassVar[list[type[Bootstrapper]]] = [
        LoadEnvironmentVariables,
        LoadConfiguration,
        RegisterProviders,
        BootProviders,
    ]

    def __init__(self, app: Application) -> None:
        self._app = app
        self._console: Console | None = None
        self._commands: list[type[Command]] = []
        self._command_paths: list[Path] = []
        self._commands_loaded: bool = False

    @property
    def console(self) -> Console:
        if self._console is not None:
            return self._console

        self._console = Console(self._app)
        self._console.auto_exits(False)
        self._console.catch_exceptions(False)

        return self._console

    def handle(self, input: Input, output: Output | None = None) -> int:
        try:
            self.bootstrap()
        except Exception as e:
            io = IO(
                input,
                output or StreamOutput(sys.stdout),
                output or StreamOutput(sys.stderr),
            )

            from cleo.ui.exception_trace import ExceptionTrace

            trace = ExceptionTrace(e)
            trace.render(io)

            return 1

        if output is None:
            output = StreamOutput(sys.stdout)

        try:
            return self.console.run(input, output, output)
        except Exception as e:
            handler = self._app.container.get(ExceptionHandler)  # type: ignore[type-abstract]

            handler.report(e)
            handler.render_for_console(output, e)

            return 1

    def bootstrap(self) -> None:
        if not self._app.has_been_bootstrapped():
            self._app.bootstrap_with(self._bootstrappers)

        if not self._commands_loaded:
            self._discover_commands()

            self._commands_loaded = True

    def add_commands(self, commands: list[type[Command]]) -> None:
        self._commands.extend(commands)

    def add_command_paths(self, commands: list[Path]) -> None:
        self._command_paths.extend(commands)

    def add_command(self, command: type[Command]) -> None:
        self._commands.append(command)

    def call(
        self,
        command_name: str,
        parameters: str | None = None,
        output: Output | None = None,
    ) -> int:
        self.bootstrap()

        input = StringInput(parameters or "")
        output = output or BufferedOutput()
        command: Command = self.console.find(command_name)

        return command.run(IO(input, output, output))

    def _discover_commands(self) -> None:
        for path in self._command_paths:
            self.load_path(path)

        for command in self._commands:
            self.console.add(command())

    def load_path(self, path: Path) -> None:
        if path.is_dir():
            for filepath in path.rglob("*.py"):
                if filepath.name.startswith("_"):
                    continue

                (
                    command_name,
                    command_factory,
                ) = self._create_command_factory_from_path(filepath)
                self.console.command_loader.register_factory(
                    command_name, command_factory
                )

    def _create_command_factory_from_path(
        self, path: Path
    ) -> tuple[str, Callable[[], Command]]:
        try:
            path = path.resolve().relative_to(self._app.base_path.resolve())
        except ValueError:
            path = path.resolve().relative_to(
                Path(expanse.__file__).resolve().parent.parent
            )

        name = path.stem.removesuffix("_command")
        classname = re.sub(r"(_)+", " ", name).title().replace(" ", "")
        command_name = " ".join(name.split("_"))

        def factory() -> Command:
            components = [
                # Import path
                path.with_suffix("").as_posix().replace("/", "."),
                # Class name
                f"{classname}Command",
            ]
            class_full_name = ".".join(components)
            class_: type[Command] = string_to_class(class_full_name)

            return self._app.container.call(class_)

        return command_name, factory



from __future__ import annotations


class PlaceholderPath:
    def __init__(self, app_path: str, relative_path: str) -> None:
        self.app_path = app_path
        self.relative_path = relative_path


def path(relative_path: str | None = None) -> PlaceholderPath:
    """
    Returns a placeholder Path that will be resolved relative to the application
    when it is bootstrapped.
    """
    return PlaceholderPath("base", relative_path or "")


def static_path(relative_path: str | None = None) -> PlaceholderPath:
    """
    Returns a placeholder Path that will be resolved relative to the application
    when it is bootstrapped.
    """
    return PlaceholderPath("static", relative_path or "")


def resource_path(relative_path: str) -> PlaceholderPath:
    """
    Returns a placeholder Path that will be resolved relative to the application
    when it is bootstrapped.
    """
    return PlaceholderPath("resources", relative_path)



from typing import Annotated

from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import MappedAsDataclass
from sqlalchemy.orm import mapped_column


primary_key = Annotated[
    int, mapped_column(init=False, primary_key=True, autoincrement=True)
]


class Model(MappedAsDataclass, DeclarativeBase): ...



from typing import ClassVar

from cleo.helpers import option
from cleo.io.inputs.argument import Argument
from cleo.io.inputs.option import Option

from expanse.common.database.migration.migrator import Migrator
from expanse.database.console.command import MigrationCommand


class DbRollbackCommand(MigrationCommand):
    name: str = "db rollback"

    description: str = "Rollback the last database migration."

    arguments: ClassVar[list[Argument]] = []
    options: ClassVar[list[Option]] = [
        option("step", None, "The number of migrations to revert.", flag=False),
    ]

    help: str = """Rollback the last database migration.

A revision can be provided to rollback to a specific revision.

<info>db rollback my-revision</info>

Alternatively, a negative integer can be provided to rollback a specific number of revisions. For instance:

<info>db rollback -2</info>

This will rollback the last two migrations.
"""

    def handle(self, migrator: Migrator) -> int:
        self.line("")

        revision = "-1"
        if self.option("step"):
            revision = f'-{self.option("step").removeprefix("-")}'

        migrator.rollback(revision=revision, io=self._io)

        return 0



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING
from typing import Any
from typing import Self

from baize.asgi import empty_receive
from baize.asgi import empty_send
from baize.asgi.requests import Request as BaseRequest

from expanse.common.http.accept_header import AcceptHeader
from expanse.common.http.url import URL


if TYPE_CHECKING:
    from expanse.asynchronous.routing.route import Route
    from expanse.asynchronous.types import Receive
    from expanse.asynchronous.types import Scope
    from expanse.asynchronous.types import Send


class Request(BaseRequest):
    def __init__(
        self, scope: Scope, receive: Receive = empty_receive, send: Send = empty_send
    ):
        super().__init__(scope=scope, receive=receive, send=send)

        self._route: Route | None = None

    @cached_property
    def url(self) -> URL:  # type: ignore[override]
        return URL(scope=self._scope)

    @cached_property
    def host(self) -> str:
        client = self.client

        if not client:
            return ""

        return client.host or ""

    @cached_property
    def acceptable_content_types(self) -> list[str]:
        return [
            item.value
            for item in AcceptHeader.from_string(self.headers.get("Accept", "")).all()
        ]

    @property
    def route(self) -> Route | None:
        return self._route

    def accepts_any_content_type(self) -> bool:
        """
        Determine if the current request accepts any content type.
        """
        acceptable = self.acceptable_content_types

        return len(acceptable) == 0 or acceptable[0] in ("*/*", "*")

    def is_json(self) -> bool:
        """
        Determine whether the request is sending JSON or not.
        """
        return "/json" in self.content_type.type or "+json" in self.content_type.type

    def wants_json(self) -> bool:
        """
        Determine whether the request is asking for JSON or not.
        """
        acceptable = self.acceptable_content_types

        return len(acceptable) > 0 and (
            "/json" in acceptable[0] or "+json" in acceptable[0]
        )

    def expects_json(self) -> bool:
        """
        Determine if the current request probably expects a JSON response.
        """
        return (
            self.is_ajax()
            and not self.is_pjax()
            and self.accepts_any_content_type()
            or self.wants_json()
        )

    def is_xml_http_request(self) -> bool:
        return self.headers.get("X-Requested-With") == "XMLHttpRequest"

    def is_ajax(self) -> bool:
        return self.is_xml_http_request()

    def is_pjax(self) -> bool:
        return self.headers.get("X-PJAX") == "true"

    def set_route(self, route: Route) -> Self:
        self._route = route

        return self

    @classmethod
    def create(
        cls, raw_url: str, method: str = "GET", scope: dict[str, Any] | None = None
    ) -> Request:
        default_headers = {
            b"User-Agent": b"Expanse",
            b"Accept": b"text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
            b"Accept-Language": b"en-us,en;q=0.5",
        }
        base_scope = {
            "type": "http",
            "scheme": "http",
            "server": ["localhost", 80],
            "headers": [],
            "REMOTE_ADDR": "127.0.0.1",
            "root_path": "",
            "http_version": "1.1",
            "path": "",
            "raw_path": "",
            "method": method.upper(),
            **(scope or {}),
        }

        headers = base_scope["headers"]
        header_names = {header[0] for header in headers}

        for default_header in default_headers:
            if default_header not in header_names:
                headers.append([default_header, default_headers[default_header]])

        url = URL(raw_url)

        if url.hostname is not None:
            base_scope["server"][0] = url.hostname

        if url.scheme:
            base_scope["scheme"] = url.scheme

            if url.scheme == "https":
                base_scope["server"][1] = 443
            else:
                base_scope["server"][1] = 80

        if url.port is not None:
            base_scope["server"][1] = url.port

        path = url.path
        if not path:
            path = "/"

        base_scope["path"] = path
        base_scope["raw_path"] = path

        query_string = ""
        if url.query:
            query_string = url.query

        base_scope["query_string"] = query_string

        return cls(scope=base_scope)


__all__ = ["Request"]



from collections.abc import Generator
from pathlib import Path
from typing import Any

import pytest

from expanse.core.application import Application
from expanse.database.database_manager import DatabaseManager


@pytest.fixture(autouse=True)
def setup_databases(app: Application, tmp_path: Path) -> Generator[None]:
    config: dict[str, dict[str, Any]] = app.container.get("config")

    config["database"] = {
        "default": "sqlite",
        "connections": {
            "sqlite": {"driver": "sqlite", "database": tmp_path.joinpath("db.sqlite")},
            "sqlite2": {
                "driver": "sqlite",
                "url": f"sqlite:///{tmp_path.joinpath('db2.sqlite').as_posix()}",
            },
            "postgresql": {
                "driver": "postgresql",
                "url": "postgresql://postgres:password@localhost:5432/expanse",
            },
            "postgresql_psycopg2": {
                "driver": "postgresql",
                "url": "postgresql+psycopg2://postgres:password@localhost:5432/expanse",
            },
            "postgresql_psycopg": {
                "driver": "postgresql",
                "url": "postgresql+psycopg://postgres:password@localhost:5432/expanse",
            },
            "postgresql_pg8000": {
                "driver": "postgresql",
                "url": "postgresql+pg8000://postgres:password@localhost:5432/expanse",
            },
            "test": {"driver": "sqlite", "database": tmp_path.joinpath("test.sqlite")},
        },
    }

    db = app.container.get(DatabaseManager)

    with db.connection("sqlite") as connection:
        connection.execute("CREATE TABLE IF NOT EXISTS my_table (id INTEGER)")
        connection.execute(
            """
            CREATE TABLE IF NOT EXISTS users (
                id INTEGER NOT NULL,
                first_name VARCHAR NOT NULL,
                last_name VARCHAR,
                email VARCHAR NOT NULL,
                PRIMARY KEY (id)
            );"""
        )
        connection.execute("INSERT INTO my_table (id) VALUES (:id)", {"id": "sqlite"})
        connection.execute(
            "INSERT INTO users (first_name, last_name, email) VALUES ('John', 'Doe', 'john@doe.com')"
        )
        connection.commit()

    with db.connection("sqlite2") as connection:
        connection.execute("CREATE TABLE IF NOT EXISTS my_table (id INTEGER)")
        connection.execute("INSERT INTO my_table (id) VALUES (:id)", {"id": "sqlite2"})
        connection.commit()

    with db.connection("postgresql") as connection:
        connection.execute("CREATE TABLE IF NOT EXISTS my_table (id VARCHAR)")
        connection.execute(
            "INSERT INTO my_table (id) VALUES (:id)",
            [
                {"id": "postgresql"},
                {"id": "postgresql_psycopg2"},
                {"id": "postgresql_psycopg"},
                {"id": "postgresql_pg8000"},
            ],
        )
        connection.commit()

    yield

    with db.connection("postgresql") as connection:
        connection.execute("DROP TABLE IF EXISTS my_table")



from __future__ import annotations

import os

from pathlib import Path
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from collections.abc import Callable


class ViewFinder:
    def __init__(self, paths: list[Path]) -> None:
        self._paths: list[Path] = paths

    def add_paths(self, paths: list[Path]) -> None:
        self._paths.extend(paths)

    def find(self, view: str) -> tuple[str, str, Callable[[], bool]] | None:
        view_path = Path(view)

        if not view_path.suffix:
            candidates = [
                view_path.with_suffix(".jinja2"),
                view_path.with_suffix(".html.jinja2"),
                view_path.with_suffix(".html"),
            ]
        else:
            candidates = [view_path]

        for path in self._paths:
            for candidate in candidates:
                if path.joinpath(candidate).exists():
                    return self._source(path.joinpath(candidate))

        return None

    def _source(self, view: Path) -> tuple[str, str, Callable[[], bool]]:
        mtime = os.path.getmtime(view)

        def uptodate() -> bool:
            try:
                return os.path.getmtime(view) == mtime
            except OSError:
                return False

        return view.read_text(), view.as_posix(), uptodate



from collections.abc import Awaitable
from collections.abc import Callable
from functools import wraps
from typing import Self

from expanse.asynchronous.container.container import Container
from expanse.asynchronous.http.request import Request
from expanse.asynchronous.http.response import Response
from expanse.asynchronous.types.http.middleware import RequestHandler


class Pipeline:
    """
    A routing pipeline are composed of layers, like middleware, through which
    a request — and corresponding response — must go through.
    """

    def __init__(self, container: Container) -> None:
        self._container = container
        self._pipes: list[Callable[[Request, RequestHandler], Awaitable[Response]]] = []
        self._request: Request | None = None

    def use(
        self, pipes: list[Callable[[Request, RequestHandler], Awaitable[Response]]]
    ) -> Self:
        self._pipes = pipes

        return self

    def send(self, request: Request) -> Self:
        self._request = request

        return self

    async def to(self, handler: RequestHandler) -> Response:
        if self._request is None:
            raise RuntimeError("No request has been set on the pipeline.")

        from expanse.asynchronous.core.helpers import _set_container

        _set_container(self._container)

        try:
            pipeline = await self._build_pipeline(handler)
            return await pipeline(self._request)
        except Exception as e:
            from expanse.asynchronous.contracts.debug.exception_handler import (
                ExceptionHandler,
            )

            if not self._container.has(ExceptionHandler):
                raise e

            exception_handler = await self._container.get(ExceptionHandler)

            await exception_handler.report(e)

            return await exception_handler.render(self._request, e)
        finally:
            _set_container(None)

    async def _build_pipeline(self, handler: RequestHandler) -> RequestHandler:
        stack = handler

        for pipe in self._pipes[::-1]:
            stack = await self._wrap(pipe)(stack)

        return stack

    def _wrap(
        self, pipe: Callable[[Request, RequestHandler], Awaitable[Response]]
    ) -> Callable[[RequestHandler], Awaitable[RequestHandler]]:
        @wraps(pipe)
        async def decorator(
            next_call: RequestHandler,
        ) -> Callable[[Request], Awaitable[Response]]:
            @wraps(next_call)
            async def handler(request: Request) -> Response:
                return await pipe(request, next_call)

            return handler

        return decorator



from typing import Any
from typing import ClassVar

from cleo.helpers import option
from cleo.io.inputs.option import Option

from expanse.asynchronous.console.commands.command import Command
from expanse.asynchronous.core.application import Application


class ServeCommand(Command):
    name: str = "serve"

    options: ClassVar[list[Option]] = [
        option("--port", flag=False, default=8000),
        option("watch"),
    ]

    async def handle(self, app: Application) -> int:
        import uvicorn

        log_level: str = "info"
        if self._io.is_verbose():
            log_level = "debug"
        elif self._io.output.is_quiet():
            log_level = "error"

        parameters: dict[str, Any] = {
            "port": int(self.option("port")),
            "log_level": log_level,
            "reload": self.option("watch"),
            "use_colors": self._io.is_decorated(),
        }

        if self.option("watch"):
            parameters["reload_dirs"] = [
                app.base_path,
                app.config_path,
                app.base_path.joinpath("routes"),
            ]

        config = uvicorn.Config("app.app:app", **parameters)
        server = uvicorn.Server(config)

        if self.option("watch"):
            from uvicorn.supervisors import ChangeReload

            sock = config.bind_socket()
            ChangeReload(config, target=server.run, sockets=[sock]).run()
        else:
            await server.serve()

        return 0



from collections.abc import Mapping
from collections.abc import MutableMapping
from os import PathLike
from typing import Any
from typing import NoReturn

from baize.wsgi import FileResponse
from baize.wsgi import JSONResponse
from baize.wsgi import PlainTextResponse

from expanse.common.core.http.exceptions import HTTPException
from expanse.http.response import Response
from expanse.routing.redirect import Redirect
from expanse.view.view_factory import ViewFactory


class Responder:
    def __init__(self, view: ViewFactory, redirect: Redirect) -> None:
        self._view = view
        self._redirect = redirect

    def redirect(self) -> Redirect:
        return self._redirect

    def text(
        self,
        content: str = "",
        status_code: int = 200,
        headers: Mapping[str, Any] | None = None,
        media_type: str = "text/plain",
    ) -> Response:
        return Response(
            response=PlainTextResponse(
                content, status_code=status_code, headers=headers, media_type=media_type
            )
        )

    def html(
        self,
        content: str = "",
        status_code: int = 200,
        headers: Mapping[str, Any] | None = None,
    ) -> Response:
        return self.text(
            content, status_code=status_code, headers=headers, media_type="text/html"
        )

    def json(
        self,
        content: Any = "",
        status_code: int = 200,
        headers: Mapping[str, Any] | None = None,
        **kwargs: Any,
    ) -> Response:
        return Response(
            response=JSONResponse(
                content, status_code=status_code, headers=headers, **kwargs
            )
        )

    def file(
        self,
        path: str | PathLike[str],
        headers: Mapping[str, str] | None = None,
        media_type: str | None = None,
        filename: str | None = None,
        chunk_size: int = 4096 * 64,
    ) -> Response:
        return Response(
            response=FileResponse(
                str(path),
                headers=headers,
                content_type=media_type,
                download_name=filename,
                chunk_size=chunk_size,
            )
        )

    def view(
        self,
        view: str,
        data: Mapping[str, Any] | None = None,
        status_code: int = 200,
        headers: MutableMapping[str, Any] | None = None,
    ) -> Response:
        return self._view.render(self._view.make(view, data, status_code, headers))

    def abort(
        self,
        status_code: int,
        message: str | None = None,
        headers: MutableMapping[str, str] | None = None,
    ) -> NoReturn:
        raise HTTPException(status_code, detail=message, headers=headers)



from collections.abc import Generator
from typing import Any
from typing import Self
from typing import TypeVar
from typing import cast
from typing import overload

from sqlalchemy import CursorResult
from sqlalchemy import Executable
from sqlalchemy import ScalarResult
from sqlalchemy import text
from sqlalchemy.engine.interfaces import CoreExecuteOptionsParameter
from sqlalchemy.engine.interfaces import _CoreAnyExecuteParams
from sqlalchemy.engine.interfaces import _CoreSingleExecuteParams
from sqlalchemy.ext.asyncio import AsyncConnection as BaseAsyncConnection
from sqlalchemy.sql.selectable import TypedReturnsRows


_T = TypeVar("_T", bound=Any)


class Connection(BaseAsyncConnection):
    @overload
    async def scalar(
        self,
        statement: TypedReturnsRows[tuple[_T]],
        parameters: _CoreSingleExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> _T | None: ...

    @overload
    async def scalar(
        self,
        statement: Executable,
        parameters: _CoreSingleExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> Any: ...

    @overload
    async def scalar(
        self,
        statement: str,
        parameters: _CoreSingleExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> Any: ...

    async def scalar(
        self,
        statement: TypedReturnsRows[tuple[_T]] | Executable | str,
        parameters: _CoreSingleExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> Any:
        if isinstance(statement, str):
            statement = text(statement)

        return await super().scalar(
            statement, parameters, execution_options=execution_options
        )

    @overload
    async def scalars(
        self,
        statement: TypedReturnsRows[tuple[_T]],
        parameters: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> ScalarResult[_T]: ...

    @overload
    async def scalars(
        self,
        statement: Executable,
        parameters: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> ScalarResult[Any]: ...

    @overload
    async def scalars(
        self,
        statement: str,
        parameters: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> ScalarResult[Any]: ...

    async def scalars(
        self,
        statement: TypedReturnsRows[tuple[_T]] | Executable | str,
        parameters: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> ScalarResult[Any]:
        if isinstance(statement, str):
            statement = text(statement)

        return await super().scalars(
            statement, parameters, execution_options=execution_options
        )

    @overload
    async def execute(
        self,
        statement: TypedReturnsRows[_T],
        parameters: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> CursorResult[_T]: ...

    @overload
    async def execute(
        self,
        statement: Executable,
        parameters: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> CursorResult[Any]: ...

    @overload
    async def execute(
        self,
        statement: str,
        parameters: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> CursorResult[Any]: ...

    async def execute(
        self,
        statement: TypedReturnsRows[_T] | Executable | str,
        parameters: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> CursorResult[Any]:
        if isinstance(statement, str):
            statement = text(statement)

        return await super().execute(
            statement, parameters, execution_options=execution_options
        )

    def __aenter__(self) -> Self:  # type: ignore[override]
        return self

    def __await__(self) -> Generator[Any, None, Self]:
        return cast(Generator[Any, None, Self], super().__await__())



from typing import Self

from expanse.asynchronous.support.service_provider import ServiceProvider
from expanse.common.support.service_providers_list import (
    ServiceProvidersList as BaseServiceProvidersList,
)


class ServiceProvidersList(BaseServiceProvidersList[ServiceProvider]):
    @classmethod
    def default(cls) -> Self:
        return cls(
            [
                "expanse.asynchronous.core.support.providers.core_service_provider.CoreServiceProvider",
                "expanse.asynchronous.core.console.providers.command_service_provider.CommandServiceProvider",
                "expanse.asynchronous.database.database_service_provider.DatabaseServiceProvider",
                "expanse.asynchronous.view.view_service_provider.ViewServiceProvider",
                "expanse.asynchronous.static.static_service_provider.StaticServiceProvider",
                "expanse.asynchronous.encryption.encryption_service_provider.EncryptionServiceProvider",
            ]
        )



import typing

from collections.abc import MutableMapping
from urllib.parse import SplitResult
from urllib.parse import urlsplit

from expanse.common.http.url_path import URLPath
from expanse.common.support._utils import string_matches


class URL:
    __slots__ = ("_url", "_components")

    def __init__(
        self,
        url: str = "",
        *,
        scope: MutableMapping[str, typing.Any] | None = None,
        environ: MutableMapping[str, typing.Any] | None = None,
        **components: typing.Any,
    ) -> None:
        if components:
            assert not url, 'Cannot set both "url" and "**components".'
            url = URL("").replace(**components).components.geturl()
        elif scope is not None:
            scheme = scope.get("scheme", "http")
            server = scope.get("server", None)
            path = scope.get("root_path", "") + scope["path"]
            query_string = scope.get("query_string", b"")

            host_header = None
            for key, value in scope["headers"]:
                if key == b"host":
                    host_header = value.decode("latin-1")
                    break
            url = self._build_url(scheme, path, query_string, server, host_header)
        elif environ is not None:
            scheme = environ["wsgi.url_scheme"]
            server = (environ["SERVER_NAME"], int(environ["SERVER_PORT"]))
            path = (
                (environ.get("SCRIPT_NAME", "") + environ.get("PATH_INFO", ""))
                .encode("latin1")
                .decode("utf8")
            )
            query_string = environ.get("QUERY_STRING", "").encode("latin-1")
            host_header = environ.get("HTTP_HOST", None)
            url = self._build_url(scheme, path, query_string, server, host_header)

        self._url = url
        self._components = urlsplit(url)

    def _build_url(
        self,
        scheme: str,
        path: str,
        query_string: bytes | str = b"",
        server: tuple[str, int | None] | None = None,
        host_header: str | None = None,
    ) -> str:
        if host_header is not None:
            url = f"{scheme}://{host_header}{path}"
        elif server is None:
            url = path
        else:
            host, port = server
            default_port = {"http": 80, "https": 443, "ws": 80, "wss": 443}[scheme]
            if port == default_port or port is None:
                url = f"{scheme}://{host}{path}"
            else:
                url = f"{scheme}://{host}:{port}{path}"

        if query_string:
            query_string = (
                query_string.decode()
                if isinstance(query_string, bytes)
                else query_string
            )
            url = f"{url}?{query_string}"

        return url

    @property
    def components(self) -> SplitResult:
        return self._components

    @property
    def full(self) -> str:
        return self._url

    @property
    def scheme(self) -> str:
        return self._components.scheme

    @property
    def netloc(self) -> str:
        return self._components.netloc

    @property
    def path(self) -> str:
        p = URLPath(self._components.path)

        return p

    @property
    def query(self) -> str:
        return self._components.query

    @property
    def fragment(self) -> str:
        return self._components.fragment

    @property
    def username(self) -> str | None:
        return self._components.username

    @property
    def password(self) -> str | None:
        return self._components.password

    @property
    def hostname(self) -> str | None:
        return self._components.hostname

    @property
    def port(self) -> int | None:
        return self._components.port

    def is_secure(self) -> bool:
        return self.scheme == "https"

    def is_(self, pattern: str | list[str]) -> bool:
        """
        Determine if the full URL matches a given pattern.
        """
        return string_matches(self._url, pattern)

    def path_is(self, pattern: str | list[str]) -> bool:
        """
        Determine if the full URL matches a given pattern.
        """
        return string_matches(self.path.lstrip("/"), pattern)

    def replace(self, **kwargs: typing.Any) -> "URL":
        if (
            "username" in kwargs
            or "password" in kwargs
            or "hostname" in kwargs
            or "port" in kwargs
        ):
            hostname = kwargs.pop("hostname", None)
            port = kwargs.pop("port", self.port)
            username = kwargs.pop("username", self.username)
            password = kwargs.pop("password", self.password)

            if hostname is None:
                netloc = self.netloc
                _, _, hostname = netloc.rpartition("@")

                if hostname[-1] != "]":
                    hostname = hostname.rsplit(":", 1)[0]

            netloc = hostname
            if port is not None:
                netloc += f":{port}"
            if username is not None:
                userpass = username
                if password is not None:
                    userpass += f":{password}"
                netloc = f"{userpass}@{netloc}"

            kwargs["netloc"] = netloc

        components = self._components._replace(**kwargs)
        return self.__class__(components.geturl())

    def __eq__(self, other: typing.Any) -> bool:
        return str(self) == str(other)

    def __str__(self) -> str:
        return self._url

    def __repr__(self) -> str:
        url = str(self)
        if self.password:
            url = str(self.replace(password="********"))
        return f"{self.__class__.__name__}({url!r})"



from __future__ import annotations

from pydantic import Field
from pydantic_settings import BaseSettings
from pydantic_settings import SettingsConfigDict

from expanse.asynchronous.support.service_providers_list import ServiceProvidersList


class Config(BaseSettings):
    name: str = "Expanse"

    # Application environment
    #
    # The environment the application is running in.
    env: str = "production"

    # Debug mode
    #
    # The debug mode mostly controls how the application behaves when an error
    # occurs. When debug mode is enabled, the application will display detailed
    # messages about the error, including a stack trace. When debug mode is disabled,
    # the application will display a generic error message instead.
    debug: bool = False

    # Encryption key
    #
    # This key is used by the application for encryption and should be set
    # to a random, 32-character string or a base64-encoded 32 bytes prefixed with `base64:`.
    # This must be set prior to deploying the application.
    secret_key: str = ""
    # The cipher method used for encryption. Currently, only aes-256-gcm is supported.
    cipher: str = "aes-256-gcm"

    # Service providers
    #
    # You can list here the service providers that you want to register automatically
    # in your application.
    providers: list[str] = Field(
        default=(
            ServiceProvidersList.default()
            .merge(
                [
                    # Package-provided providers
                ]
            )
            .merge(
                [
                    # Application service providers
                    # "app.providers.app_service_provider.AppServiceProvider",
                ]
            )
            .to_list()
        )
    )

    model_config = SettingsConfigDict(env_prefix="app_")



from pathlib import Path

from pydantic_settings import BaseSettings
from pydantic_settings import SettingsConfigDict


class Config(BaseSettings):
    database: Path = Path("database")
    resources: Path = Path("resources")
    static: Path = Path("static")

    model_config = SettingsConfigDict(env_prefix="path_")



from __future__ import annotations

from collections.abc import Awaitable
from collections.abc import Callable
from typing import Any
from typing import TypeAlias


Endpoint: TypeAlias = Callable[..., Awaitable[Any]] | Callable[..., Any]



from pathlib import Path
from typing import Any

from expanse.common.core.application import Application


class Stub:
    def __init__(self, path: Path) -> None:
        self._path: Path = path

    def generate(self, parameters: dict[str, Any]) -> str:
        content = self._path.read_text()

        return content.format(**parameters)

    def generate_to(self, path: Path, parameters: dict[str, Any]) -> None:
        path.write_text(self.generate(parameters))


class Generator:
    def __init__(self, app: Application) -> None:
        self._app: Application = app

    def stub(self, stub: str) -> Stub:
        path = self.resolve(stub)

        if path is None:
            raise ValueError(f"Stub not found: {stub}")

        return Stub(path)

    def resolve(self, stub: str) -> Path | None:
        paths: list[Path] = [
            self._app.path("stubs"),
            Path(__file__).parent.joinpath("stubs"),
        ]
        for path in paths:
            stub_path = path.joinpath(stub).with_suffix(".stub")
            if stub_path.exists():
                return path.joinpath(stub_path)

        return None



# ruff: noqa: I002
import builtins
import collections
import inspect
import logging
import types
import typing

from abc import ABC
from abc import abstractmethod
from collections import defaultdict
from collections.abc import Callable
from inspect import Parameter
from typing import Annotated
from typing import Any
from typing import Self
from typing import TypedDict
from typing import TypeVar
from typing import get_args
from typing import get_origin

from expanse.common.support._utils import eval_type_lenient


T = TypeVar("T")

_builtins = {d for d in dir(builtins) if isinstance(getattr(builtins, d), type)}
_typing_builtins = {Any}
_typing_builtins_strings = {str(t) for t in _typing_builtins}
_EMPTY = object()

logger = logging.getLogger(__name__)

_Callback = Callable[..., Any]


class _Scoped(TypedDict):
    bindings: dict[str | type, Any]
    terminating_callbacks: list[_Callback]
    after_resolving_callbacks: dict[str | type, list[_Callback]]
    aliases: dict[str, str | type]


class Container(ABC):
    def __init__(self) -> None:
        self._bindings: dict[str | type, Any] = {}
        self._resolved: dict[str | type, bool] = {}
        self._instances: dict[str | type, Any] = {}
        self._aliases: dict[str, str | type] = {}

        self._scoped_bindings: dict[str | type, Any] = {}

        self._after_resolving_callbacks: dict[str | type, list[_Callback]] = (
            defaultdict(list)
        )
        self._scoped: _Scoped = {
            "bindings": {},
            "terminating_callbacks": [],
            "after_resolving_callbacks": defaultdict(list),
            "aliases": {},
        }

    def register(
        self,
        abstract: type | str,
        concrete: Any = None,
        *,
        cached: bool = False,
        scoped: bool = False,
    ) -> None:
        if concrete is None:
            concrete = abstract

        if not isinstance(concrete, types.FunctionType | types.MethodType):
            concrete = self._concrete_closure(abstract, concrete)

        if scoped:
            self._scoped["bindings"][abstract] = {
                "concrete": concrete,
                "cached": cached,
            }
        else:
            self._bindings[abstract] = {"concrete": concrete, "cached": cached}

    def singleton(
        self, abstract: type | str, concrete: Any = None, *, scoped: bool = False
    ) -> None:
        self.register(abstract, concrete, cached=True, scoped=scoped)

    def scoped(self, abstract: type | str, concrete: Any = None) -> None:
        self.singleton(abstract, concrete, scoped=True)

    def instance(self, abstract: type | str, instance: Any) -> None:
        self._instances[abstract] = instance

    def alias(self, abstract: str | type, alias: str) -> None:
        self._aliases[alias] = abstract

    def bound(self, abstract: str | type) -> bool:
        return abstract in self._bindings or abstract in self._instances

    def has(self, abstract: str | type) -> bool:
        return self.bound(abstract)

    @abstractmethod
    def create_scoped_container(self) -> Self: ...

    def has_scoped_bindings(self) -> bool:
        return bool(self._scoped["bindings"])

    def resolved(self, abstract: str | type) -> bool:
        abstract = self._get_alias(abstract)

        return abstract in self._resolved

    def after_resolving(self, abstract: str | type, callback: _Callback) -> None:
        abstract = self._get_alias(abstract)

        actual_abstract: str | type = abstract
        origin = get_origin(abstract)
        if origin is Annotated:
            actual_abstract, *_ = get_args(abstract)

        if self._is_scoped(abstract):
            self._scoped["after_resolving_callbacks"][abstract].append(callback)
        elif self._is_scoped(actual_abstract):
            self._scoped["after_resolving_callbacks"][actual_abstract].append(callback)
        elif abstract in self._bindings:
            self._after_resolving_callbacks[abstract].append(callback)
        elif actual_abstract in self._bindings:
            self._after_resolving_callbacks[actual_abstract].append(callback)
        else:
            self._after_resolving_callbacks[abstract].append(callback)

    @abstractmethod
    def _concrete_closure(
        self, abstract: str | type, concrete: Any
    ) -> Callable[[Self], Any]: ...

    def _can_build(self, abstract: str | type, concrete: Any) -> bool:
        return abstract == concrete or isinstance(
            concrete, types.FunctionType | types.MethodType
        )

    def _is_cached(self, abstract: str | type) -> bool:
        return abstract in self._instances or self._bindings.get(abstract, {}).get(
            "cached", False
        )

    def _is_scoped(self, abstract: str | type) -> bool:
        return self._get_alias(abstract) in self._scoped["bindings"]

    def _mark_as_resolved(self, abstract: str | type) -> None:
        self._resolved[abstract] = True

    def _get_class(
        self, parameter: Parameter, *, _globals: dict[str, Any] | None = None
    ) -> type | None:
        type_ = parameter.annotation

        if type_ is Parameter.empty:
            return None

        # TODO: handle optionals

        if isinstance(type_, types.UnionType):
            # TODO: check that the union type is a single type optional
            # Get the first type of the type union
            type_ = get_args(type_)[0]

        if self._is_builtin(type_, _globals=_globals):
            return None

        return type_

    def _is_builtin(
        self, type_: type, *, _globals: dict[str, Any] | None = None
    ) -> bool:
        if isinstance(type_, str):
            type_ = eval_type_lenient(type_, _globals, _globals)

            if isinstance(type_, typing.ForwardRef):
                type_ = type_.__forward_arg__

                if type_ in _typing_builtins_strings:
                    return True

                return False

        module = inspect.getmodule(type_)
        if module == builtins:
            return True

        if type_ in _typing_builtins:
            return True

        if (
            module == typing or module == collections.abc
        ) and type_.__name__ == "Callable":
            return True

        return False

    def _get_alias(self, abstract: str | type) -> str | type:
        if not isinstance(abstract, str):
            return abstract

        return self._aliases.get(abstract, abstract)

    def _is_lambda(self, callable: _Callback) -> bool:
        return (
            isinstance(callable, types.FunctionType) and callable.__name__ == "<lambda>"
        )



import functools

from collections.abc import Callable
from typing import ParamSpec
from typing import TypeVar

import anyio.to_thread


P = ParamSpec("P")
T = TypeVar("T")


async def run_in_threadpool(
    func: Callable[P, T], *args: P.args, **kwargs: P.kwargs
) -> T:
    if kwargs:  # pragma: no cover
        # run_sync doesn't accept 'kwargs', so bind them in here
        func = functools.partial(func, **kwargs)
    return await anyio.to_thread.run_sync(func, *args)



from pathlib import Path

from expanse.asynchronous.core.application import Application
from expanse.asynchronous.encryption.console.commands.encryption_key_generate import (
    EncryptionKeyGenerateCommand,  # noqa: F401
)
from expanse.asynchronous.testing.command_tester import CommandTester


async def test_generate_key_without_env_file(
    command_tester: CommandTester, app: Application, tmp_path: Path
) -> None:
    app._environment_path = tmp_path

    command = command_tester.command("encryption key generate")

    return_code = await command.run()
    assert return_code == 1

    expected = """The .env file does not exist.
"""

    assert command.output.fetch() == expected


async def test_generate_key_with_env_file_but_without_app_secret_key(
    command_tester: CommandTester, app: Application, tmp_path: Path
) -> None:
    tmp_path.joinpath(".env").write_text("")
    app._environment_path = tmp_path

    command = command_tester.command("encryption key generate")

    return_code = await command.run()
    assert return_code == 1

    expected = """The application key could not be set. \
Please check your .env file and that a APP_SECRET_KEY is present.
"""

    assert command.output.fetch() == expected


async def test_generate_key_with_env_file_and_with_empty_app_secret_key(
    command_tester: CommandTester, app: Application, tmp_path: Path
) -> None:
    app.config["app.secret_key"] = ""
    tmp_path.joinpath(".env").write_text("APP_SECRET_KEY=")
    app._environment_path = tmp_path

    command = command_tester.command("encryption key generate")

    return_code = await command.run()
    assert return_code == 0

    expected = """Application secret key generated successfully.
"""

    assert command.output.fetch() == expected

    assert app.config["app.secret_key"] != ""
    assert (
        tmp_path.joinpath(".env").read_text().strip()
        == f"APP_SECRET_KEY={app.config['app.secret_key']}"
    )


async def test_generate_key_with_env_file_and_with_non_empty_app_secret_key(
    command_tester: CommandTester, app: Application, tmp_path: Path
) -> None:
    app.config["app.secret_key"] = "S" * 32
    tmp_path.joinpath(".env").write_text(f"APP_SECRET_KEY={'S' * 32}")
    app._environment_path = tmp_path

    command = command_tester.command("encryption key generate")

    return_code = await command.with_user_input("y\n").run()
    assert return_code == 0

    expected = """Do you want to replace the existing application secret key? (yes/no) [no]\
 Application secret key generated successfully.
"""

    assert command.output.fetch() == expected

    assert app.config["app.secret_key"] != "S" * 32
    assert (
        tmp_path.joinpath(".env").read_text().strip()
        == f"APP_SECRET_KEY={app.config['app.secret_key']}"
    )


async def test_generate_key_with_env_file_and_with_non_empty_app_secret_key_and_no_confirmation(
    command_tester: CommandTester, app: Application, tmp_path: Path
) -> None:
    app.config["app.secret_key"] = "S" * 32
    tmp_path.joinpath(".env").write_text(f"APP_SECRET_KEY={'S' * 32}")
    app._environment_path = tmp_path

    command = command_tester.command("encryption key generate")

    return_code = await command.with_user_input("n\n").run()
    assert return_code == 1

    expected = (
        "Do you want to replace the existing application secret key? (yes/no) [no] "
    )

    assert command.output.fetch() == expected

    assert app.config["app.secret_key"] == "S" * 32
    assert tmp_path.joinpath(".env").read_text().strip() == f"APP_SECRET_KEY={'S' * 32}"


async def test_generate_key_show_only(
    command_tester: CommandTester, app: Application, tmp_path: Path
) -> None:
    command = command_tester.command("encryption key generate")

    return_code = await command.run("--show")
    assert return_code == 0

    assert command.output.fetch().strip().startswith("base64:")



import math
import re
import typing
import urllib.parse
import uuid

from dataclasses import dataclass
from re import Pattern

from expanse.asynchronous.http.request import Request as ASGIRequest
from expanse.common.http.url_path import URLPath
from expanse.common.routing.exceptions import NotEnoughURLParameters
from expanse.common.routing.route import Match
from expanse.common.routing.route import Route
from expanse.http.request import Request


T = typing.TypeVar("T")


class Convertor(typing.Generic[T]):
    regex: typing.ClassVar[str] = ""

    def convert(self, value: str) -> T:
        raise NotImplementedError()  # pragma: no cover

    def to_string(self, value: T) -> str:
        raise NotImplementedError()  # pragma: no cover


class StringConvertor(Convertor):
    regex = "[^/]+"

    def convert(self, value: str) -> str:
        return value

    def to_string(self, value: str) -> str:
        value = str(value)
        assert "/" not in value, "May not contain path separators"
        assert value, "Must not be empty"
        return value


class PathConvertor(Convertor):
    regex = ".*"

    def convert(self, value: str) -> str:
        return str(value)

    def to_string(self, value: str) -> str:
        return str(value)


class IntegerConvertor(Convertor):
    regex = "[0-9]+"

    def convert(self, value: str) -> int:
        return int(value)

    def to_string(self, value: int) -> str:
        value = int(value)
        assert value >= 0, "Negative integers are not supported"
        return str(value)


class FloatConvertor(Convertor):
    regex = r"[0-9]+(\.[0-9]+)?"

    def convert(self, value: str) -> float:
        return float(value)

    def to_string(self, value: float) -> str:
        value = float(value)
        assert value >= 0.0, "Negative floats are not supported"
        assert not math.isnan(value), "NaN values are not supported"
        assert not math.isinf(value), "Infinite values are not supported"
        return f"{value:0.20f}".rstrip("0").rstrip(".")


class UUIDConvertor(Convertor):
    regex = "[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"

    def convert(self, value: str) -> uuid.UUID:
        return uuid.UUID(value)

    def to_string(self, value: uuid.UUID) -> str:
        return str(value)


@dataclass
class MatchResult:
    result: Match
    format: str
    convertor: Convertor | None

    def convert(self):
        pass


# Match parameters in URL paths, eg. '{param}', and '{param:int}'
PARAM_REGEX = re.compile("{([a-zA-Z_][a-zA-Z0-9_]*)(:[a-zA-Z_][a-zA-Z0-9_]*)?}")


class RouteMatcher:
    def __init__(self) -> None:
        self._convertors: dict[str, Convertor] = {
            "str": StringConvertor(),
            "path": PathConvertor(),
            "int": IntegerConvertor(),
            "float": FloatConvertor(),
            "uuid": UUIDConvertor(),
        }
        self._compiled: dict[str, tuple[Pattern, str, dict[str, Convertor]]] = {}

    def match(self, route: Route, request: Request | ASGIRequest) -> Match:
        match request:
            case Request():
                return self._match_from_request(route, request)

        return self._match_from_asgi_request(route, request)

    def compile_path(self, path: str) -> tuple[Pattern, str, dict[str, Convertor]]:
        """
        Given a path string, like: "/{username:str}",
        or a host string, like: "{subdomain}.mydomain.org", return a three-tuple
        of (regex, format, {param_name:convertor}).

        regex:      "/(?P<username>[^/]+)"
        format:     "/{username}"
        convertors: {"username": StringConvertor()}
        """
        if path in self._compiled:
            return self._compiled[path]

        is_host = not path.startswith("/")

        path_regex = "^"
        path_format = ""
        duplicated_params = set()

        idx = 0
        param_convertors = {}
        for match in PARAM_REGEX.finditer(path):
            param_name, convertor_type = match.groups("str")
            convertor_type = convertor_type.lstrip(":")

            if convertor_type not in self._convertors:
                raise ValueError(f"Unknown parameter convertor {convertor_type}.")

            convertor = self._convertors[convertor_type]

            path_regex += re.escape(path[idx : match.start()])
            path_regex += f"(?P<{param_name}>{convertor.regex})"

            path_format += path[idx : match.start()]
            path_format += f"{{{param_name}}}"

            if param_name in param_convertors:
                duplicated_params.add(param_name)

            param_convertors[param_name] = convertor

            idx = match.end()

        if duplicated_params:
            names = ", ".join(sorted(duplicated_params))
            ending = "s" if len(duplicated_params) > 1 else ""
            raise ValueError(f"Duplicated param name{ending} {names} at path {path}")

        if is_host:
            # Align with `Host.matches()` behavior, which ignores port.
            hostname = path[idx:].split(":")[0]
            path_regex += re.escape(hostname) + "$"
        else:
            path_regex += re.escape(path[idx:]) + "$"

        path_format += path[idx:]

        self._compiled[path] = re.compile(path_regex), path_format, param_convertors

        return self._compiled[path]

    def add_convertor(self, key: str, convertor: Convertor) -> None:
        self._convertors[key] = convertor

    def url(self, path: str, /, **parameters) -> URLPath:
        path_regex, path_format, convertors = self.compile_path(path)

        param_names = set(parameters.keys())
        expected_param_names = set(convertors.keys())

        if not expected_param_names.issubset(param_names):
            raise NotEnoughURLParameters(
                f"Not enough parameters for URL {path}: "
                f"missing "
                f"{', '.join(sorted(expected_param_names.difference(param_names)))}"
            )

        for key, value in list(parameters.items()):
            if "{" + key + "}" in path_format:
                convertor = convertors[key]
                value = convertor.to_string(value)
                path_format = path_format.replace("{" + key + "}", value)
                parameters.pop(key)

        if parameters:
            # If we still have parameters, convert them to a query string
            path_format += f"?{urllib.parse.urlencode(parameters)}"

        return URLPath(path_format)

    def _match_from_request(self, route: Route, request: Request) -> Match:
        path_regex, fmt, convertors = self.compile_path(route.path)
        match = path_regex.match(request.url.path)
        if match:
            matched_params = match.groupdict()
            for key, value in matched_params.items():
                matched_params[key] = convertors[key].convert(value)
            path_params = request.path_params
            path_params.update(matched_params)

            if route.methods and request.method not in route.methods:
                return Match.PARTIAL

            return Match.FULL

        return Match.NONE

    def _match_from_asgi_request(self, route: Route, request: ASGIRequest) -> Match:
        path_regex, fmt, convertors = self.compile_path(route.path)
        match = path_regex.match(request.url.path)
        if match:
            matched_params = match.groupdict()
            for key, value in matched_params.items():
                matched_params[key] = convertors[key].convert(value)
            path_params = request.path_params
            path_params.update(matched_params)

            if route.methods and request.method not in route.methods:
                return Match.PARTIAL

            return Match.FULL

        return Match.NONE



from expanse.asynchronous.routing.router import Router
from expanse.asynchronous.testing.command_tester import CommandTester


async def post_endpoint() -> str:
    return ""


async def delete_endpoint() -> str:
    return ""


async def test_route_listing(command_tester: CommandTester, router: Router) -> None:
    router.post("/post", post_endpoint)
    with router.group("group", prefix="/group") as group:
        group.delete("/delete", delete_endpoint, name="delete_name")

    command = command_tester.command("routes list")

    assert await command.run() == 0

    expected_output = """
  DELETE    /group/delete (group.delete_name)
  POST      /post
  GET|HEAD  /static/{path:path} (static)
"""
    assert command.output.fetch() == expected_output


async def test_route_listing_verbose(
    command_tester: CommandTester, router: Router
) -> None:
    router.post("/post", post_endpoint)
    with router.group("group", prefix="/group") as group:
        group.delete("/delete", delete_endpoint, name="delete_name")

    command = command_tester.command("routes list")

    assert await command.run("-v") == 0

    expected_output = f"""
  DELETE    /group/delete (group.delete_name)
            {delete_endpoint.__module__}.delete_endpoint
  POST      /post
            {post_endpoint.__module__}.post_endpoint
  GET|HEAD  /static/{{path:path}} (static)
            expanse.asynchronous.static.static.Static.get
"""
    assert command.output.fetch() == expected_output



from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from expanse.common.http.url import URL


class URLPath(str):
    def __new__(cls, path: str, protocol: str = "", host: str = "") -> URLPath:
        assert protocol in ("http", "websocket", "")
        return str.__new__(cls, path)

    def __init__(self, path: str, protocol: str = "", host: str = "") -> None:
        self.protocol = protocol
        self.host = host

    def make_absolute_url(self, base_url: str | URL) -> URL:
        from expanse.common.http.url import URL

        if isinstance(base_url, str):
            base_url = URL(base_url)

        if self.protocol:
            scheme = {
                "http": {True: "https", False: "http"},
                "websocket": {True: "wss", False: "ws"},
            }[self.protocol][base_url.is_secure()]

        else:
            scheme = base_url.scheme

        netloc = self.host or base_url.netloc
        path = base_url.path.rstrip("/") + str(self)
        return URL(scheme=scheme, netloc=netloc, path=path)



from typing import ClassVar

from cleo.helpers import option
from cleo.io.inputs.argument import Argument
from cleo.io.inputs.option import Option

from expanse.asynchronous.database.console.command import MigrationCommand
from expanse.asynchronous.support._concurrency import run_in_threadpool
from expanse.common.database.migration.migrator import Migrator


class DbRollbackCommand(MigrationCommand):
    name: str = "db rollback"

    description: str = "Rollback the last database migration."

    arguments: ClassVar[list[Argument]] = []
    options: ClassVar[list[Option]] = [
        option("step", None, "The number of migrations to apply.", flag=False)
    ]

    help: str = """Rollback the last database migration.

A number of migrations to rollback can be specified with the --step option.

<info>db rollback --step 2</info>

Would rollback the last two migrations.
"""

    async def handle(self, migrator: Migrator) -> int:
        self.line("")

        revision = "-1"
        if self.option("step"):
            revision = f'-{self.option("step").removeprefix("-")}'

        await run_in_threadpool(migrator.rollback, revision=revision, io=self._io)

        return 0



from __future__ import annotations

import inspect

from typing import TYPE_CHECKING
from typing import Any
from typing import ClassVar
from typing import Generic
from typing import Literal
from typing import TypeVar
from typing import cast

from cleo.formatters.style import Style
from cleo.io.inputs.definition import Definition
from cleo.io.null_io import NullIO
from cleo.io.outputs.output import Verbosity


if TYPE_CHECKING:
    from contextlib import AbstractContextManager

    from cleo.io.inputs.argument import Argument
    from cleo.io.inputs.option import Option
    from cleo.io.io import IO
    from cleo.ui.progress_bar import ProgressBar
    from cleo.ui.progress_indicator import ProgressIndicator
    from cleo.ui.question import Question
    from cleo.ui.table import Rows
    from cleo.ui.table import Table
    from cleo.ui.table_separator import TableSeparator

    from expanse.common.console.console import Console  # noqa: TID

TConsole = TypeVar("TConsole", bound="Console")


class Command(Generic[TConsole]):
    name: str | None = None

    description: str = ""

    arguments: ClassVar[list[Argument]] = []
    options: ClassVar[list[Option]] = []
    aliases: ClassVar[list[str]] = []

    help: str = ""

    enabled: bool = True
    hidden: bool = False

    usages: ClassVar[list[str]] = []

    def __init__(self) -> None:
        self._definition = Definition()
        self._full_definition: Definition | None = None
        self._console: TConsole | None = None
        self._ignore_validation_errors = False
        self._synopsis: dict[str, str] = {}

        self.configure()

        for i, usage in enumerate(self.usages):
            if self.name and not usage.startswith(self.name):
                self.usages[i] = f"{self.name} {usage}"

        self._io: IO = NullIO()

    @property
    def console(self) -> TConsole | None:
        return self._console

    @property
    def definition(self) -> Definition:
        if self._full_definition is not None:
            return self._full_definition

        return self._definition

    @property
    def processed_help(self) -> str:
        help_text = self.help
        if not self.help:
            help_text = self.description

        is_single_command = self._console and self._console.is_single_command()

        current_script = self._console.name if self._console else inspect.stack()[-1][1]

        return help_text.format(
            command_name=self.name,
            command_full_name=current_script
            if is_single_command
            else f"{current_script} {self.name}",
            script_name=current_script,
        )

    def argument(self, name: str) -> Any:
        """
        Get the value of a command argument.
        """
        return self._io.input.argument(name)

    def option(self, name: str) -> Any:
        """
        Get the value of a command option.
        """
        return self._io.input.option(name)

    def confirm(
        self, question: str, default: bool = False, true_answer_regex: str = "(?i)^y"
    ) -> bool:
        """
        Confirm a question with the user.
        """
        from cleo.ui.confirmation_question import ConfirmationQuestion

        confirmation = ConfirmationQuestion(
            question, default=default, true_answer_regex=true_answer_regex
        )
        return cast(bool, confirmation.ask(self._io))

    def ask(self, question: str | Question, default: Any | None = None) -> Any:
        """
        Prompt the user for input.
        """
        from cleo.ui.question import Question

        if not isinstance(question, Question):
            question = Question(question, default=default)

        return question.ask(self._io)

    def secret(self, question: str | Question, default: Any | None = None) -> Any:
        """
        Prompt the user for input but hide the answer from the console.
        """
        from cleo.ui.question import Question

        if not isinstance(question, Question):
            question = Question(question, default=default)

        question.hide()

        return question.ask(self._io)

    def choice(
        self,
        question: str,
        choices: list[str],
        default: Any | None = None,
        attempts: int | None = None,
        multiple: bool = False,
    ) -> Any:
        """
        Give the user a single choice from an list of answers.
        """
        from cleo.ui.choice_question import ChoiceQuestion

        choice = ChoiceQuestion(question, choices, default)

        choice.set_max_attempts(attempts)
        choice.set_multi_select(multiple)

        return choice.ask(self._io)

    def create_question(
        self,
        question: str,
        type: Literal["choice", "confirmation"] | None = None,
        **kwargs: Any,
    ) -> Question:
        """
        Returns a Question of specified type.
        """
        from cleo.ui.choice_question import ChoiceQuestion
        from cleo.ui.confirmation_question import ConfirmationQuestion
        from cleo.ui.question import Question

        if not type:
            return Question(question, **kwargs)

        if type == "choice":
            return ChoiceQuestion(question, **kwargs)

        if type == "confirmation":
            return ConfirmationQuestion(question, **kwargs)

    def table(
        self,
        header: str | None = None,
        rows: Rows | None = None,
        style: str | None = None,
    ) -> Table:
        """
        Return a Table instance.
        """
        from cleo.ui.table import Table

        table = Table(self._io, style=style)

        if header:
            table.set_headers([header])

        if rows:
            table.set_rows(rows)

        return table

    def table_separator(self) -> TableSeparator:
        """
        Return a TableSeparator instance.
        """
        from cleo.ui.table_separator import TableSeparator

        return TableSeparator()

    def render_table(self, headers: str, rows: Rows, style: str | None = None) -> None:
        """
        Format input to textual table.
        """
        table = self.table(headers, rows, style)

        table.render()

    def write(self, text: str, style: str | None = None) -> None:
        """
        Writes a string without a new line.
        Useful if you want to use overwrite().
        """
        styled = f"<{style}>{text}</>" if style else text

        self._io.write(styled)

    def line(
        self,
        text: str,
        style: str | None = None,
        verbosity: Verbosity = Verbosity.NORMAL,
    ) -> None:
        """
        Write a string as information output.
        """
        styled = f"<{style}>{text}</>" if style else text

        self._io.write_line(styled, verbosity=verbosity)

    def line_error(
        self,
        text: str,
        style: str | None = None,
        verbosity: Verbosity = Verbosity.NORMAL,
    ) -> None:
        """
        Write a string as information output to stderr.
        """
        styled = f"<{style}>{text}</>" if style else text

        self._io.write_error_line(styled, verbosity)

    def info(self, text: str) -> None:
        """
        Write a string as information output.

        :param text: The line to write
        :type text: str
        """
        self.line(text, "info")

    def comment(self, text: str) -> None:
        """
        Write a string as comment output.

        :param text: The line to write
        :type text: str
        """
        self.line(text, "comment")

    def question(self, text: str) -> None:
        """
        Write a string as question output.

        :param text: The line to write
        :type text: str
        """
        self.line(text, "question")

    def progress_bar(self, max: int = 0) -> ProgressBar:
        """
        Creates a new progress bar
        """
        from cleo.ui.progress_bar import ProgressBar

        return ProgressBar(self._io, max=max)

    def progress_indicator(
        self,
        fmt: str | None = None,
        interval: int = 100,
        values: list[str] | None = None,
    ) -> ProgressIndicator:
        """
        Creates a new progress indicator.
        """
        from cleo.ui.progress_indicator import ProgressIndicator

        return ProgressIndicator(self._io, fmt, interval, values)

    def spin(
        self,
        start_message: str,
        end_message: str,
        fmt: str | None = None,
        interval: int = 100,
        values: list[str] | None = None,
    ) -> AbstractContextManager[ProgressIndicator]:
        """
        Automatically spin a progress indicator.
        """
        spinner = self.progress_indicator(fmt, interval, values)

        return spinner.auto(start_message, end_message)

    def add_style(
        self,
        name: str,
        fg: str | None = None,
        bg: str | None = None,
        options: list[str] | None = None,
    ) -> None:
        """
        Adds a new style
        """
        style = Style(fg, bg, options)
        self._io.output.formatter.set_style(name, style)
        self._io.error_output.formatter.set_style(name, style)

    def overwrite(self, text: str) -> None:
        """
        Overwrites the current line.

        It will not add a new line so use line('')
        if necessary.
        """
        self._io.overwrite(text)

    def configure(self) -> None:
        for argument in self.arguments:
            self._definition.add_argument(argument)

        for option in self.options:
            self._definition.add_option(option)

    def ignore_validation_errors(self) -> None:
        self._ignore_validation_errors = True

    def set_console(self, console: TConsole | None = None) -> None:
        self._console = console

        self._full_definition = None

    def merge_console_definition(self, merge_args: bool = True) -> None:
        if self._console is None:
            return

        self._full_definition = Definition()
        self._full_definition.add_options(self._definition.options)
        self._full_definition.add_options(self._console.definition.options)

        if merge_args:
            self._full_definition.set_arguments(self._console.definition.arguments)
            self._full_definition.add_arguments(self._definition.arguments)
        else:
            self._full_definition.set_arguments(self._definition.arguments)

    # Compatibility with Cleo's descriptors
    merge_application_definition = merge_console_definition

    def synopsis(self, short: bool = False) -> str:
        key = "short" if short else "long"

        if key not in self._synopsis:
            self._synopsis[key] = f"{self.name} {self.definition.synopsis(short)}"

        return self._synopsis[key]



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import Literal
from typing import Self
from typing import overload

from jinja2.environment import Environment
from jinja2.loaders import FunctionLoader

from expanse.http.response import Response
from expanse.view.view import View


if TYPE_CHECKING:
    from collections.abc import Mapping
    from collections.abc import MutableMapping

    from expanse.view.view_finder import ViewFinder


class ViewFactory:
    def __init__(self, finder: ViewFinder, debug: bool = False) -> None:
        self._finder: ViewFinder = finder

        self._env = Environment(
            loader=FunctionLoader(finder.find),
            cache_size=0 if debug else 400,
            auto_reload=debug,
        )

    def make(
        self,
        view: str,
        data: Mapping[str, Any] | None = None,
        status_code: int = 200,
        headers: MutableMapping[str, Any] | None = None,
    ) -> View:
        return View(view, data, status_code=status_code, headers=headers)

    @overload
    def render(self, view: View, raw: Literal[False] = False) -> Response: ...

    @overload
    def render(self, view: View, raw: Literal[True] = True) -> str: ...

    def render(self, view: View, raw: bool = False) -> Response | str:
        template = self._env.get_template(view.identifier)

        content = template.render(view.data)

        if raw:
            return content

        return Response(
            content=content,
            content_type="text/html",
            status_code=view.status_code,
            headers=view.headers,
        )

    def exists(self, view: str) -> bool:
        return self._finder.find(view) is not None

    def register_global(self, **globals: Any) -> Self:
        self._env.globals.update(globals)

        return self



from collections.abc import Mapping
from collections.abc import MutableMapping
from typing import Any


class View:
    def __init__(
        self,
        identifier: str,
        data: Mapping[Any, Any] | None = None,
        status_code: int = 200,
        headers: MutableMapping[str, Any] | None = None,
    ) -> None:
        self._identifier = identifier
        self._data = data or {}
        self._status_code = status_code
        self._headers = headers

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def data(self) -> Mapping[Any, Any]:
        return self._data

    @property
    def status_code(self) -> int:
        return self._status_code

    @property
    def headers(self) -> MutableMapping[str, Any] | None:
        return self._headers



from expanse.asynchronous.http.helpers import json
from expanse.asynchronous.http.response import Response
from expanse.asynchronous.routing.router import Router
from expanse.asynchronous.testing.client import TestClient
from expanse.common.http.query import Query
from tests.asynchronous.integration.http.fixtures.request.models import FooModel


async def index_validated(query: Query[FooModel]) -> Response:
    return await json({"bar": query.bar})


def test_simple_form_data_are_converted_if_validation_model(
    router: Router, client: TestClient
) -> None:
    router.post("/", index_validated)

    response = client.post("/?bar=42", data={"bar": "42"})

    assert response.json() == {"bar": 42}



from __future__ import annotations

from typing import TYPE_CHECKING

from expanse.asynchronous.routing.redirect import Redirect
from expanse.asynchronous.routing.router import Router
from expanse.asynchronous.support.service_provider import ServiceProvider


if TYPE_CHECKING:
    from expanse.asynchronous.view.view_factory import ViewFactory
    from expanse.common.http.url_path import URLPath


class RoutingServiceProvider(ServiceProvider):
    async def register(self) -> None:
        self._container.singleton(Router)
        self._container.alias(Router, "router")
        self._container.scoped(Redirect)

    async def boot(self) -> None:
        await self._container.on_resolved("view", self._register_view_globals)

    async def _register_view_globals(self, view: ViewFactory) -> None:
        router = await self._container.get(Router)

        async def route(name: str, **parameters) -> URLPath:
            return await router.route(name, parameters)

        async def url(name: str, **parameters) -> URLPath:
            return await router.url(name, parameters)

        view.register_global(url=url)
        view.register_global(route=route)



from collections.abc import AsyncGenerator
from pathlib import Path
from typing import TYPE_CHECKING

from expanse.asynchronous.core.application import Application
from expanse.asynchronous.database.connection import Connection
from expanse.asynchronous.database.database_manager import DatabaseManager
from expanse.asynchronous.database.session import Session
from expanse.asynchronous.exceptions.handler import ExceptionHandler
from expanse.asynchronous.support.service_provider import ServiceProvider
from expanse.common.configuration.config import Config
from expanse.common.database.migration.migrator import Migrator


if TYPE_CHECKING:
    from expanse.asynchronous.core.console.gateway import Gateway


class DatabaseServiceProvider(ServiceProvider):
    async def register(self) -> None:
        self._container.singleton(DatabaseManager)
        self._container.scoped(Session, self._create_session)
        self._container.scoped(Connection, self._create_connection)
        self._container.singleton(Migrator, self._create_migrator)
        await self._container.on_resolved(
            ExceptionHandler, self._configure_exception_handler
        )

    async def boot(self) -> None:
        from expanse.asynchronous.core.console.gateway import Gateway

        await self._container.on_resolved(Gateway, self._register_command_path)

    async def _create_connection(
        self, db: DatabaseManager, name: str | None = None
    ) -> AsyncGenerator[Connection]:
        connection = db.connection(name)

        await connection.start()

        yield connection

        await connection.close()

    async def _create_session(
        self, db: DatabaseManager, name: str | None = None
    ) -> AsyncGenerator[Session]:
        session = db.session(name)

        yield session

        await session.close()

    async def _register_command_path(self, gateway: "Gateway") -> None:
        await gateway.load_path(Path(__file__).parent.joinpath("console/commands"))

    async def _create_migrator(self, app: Application, config: Config) -> Migrator:
        migrator = Migrator(app)

        # We want to build a synchronous database manager that can be safely
        # used in migrations and in env.py Alembic file.
        from expanse.core.application import Application as SyncApplication
        from expanse.database.database_manager import (
            DatabaseManager as SyncDatabaseManager,
        )

        sync_app = SyncApplication(app.base_path)
        sync_app.set_config(config)

        migrator.config.attributes["db"] = SyncDatabaseManager(sync_app)

        return migrator

    async def _configure_exception_handler(self, handler: ExceptionHandler) -> None:
        from sqlalchemy.exc import NoResultFound

        from expanse.common.core.http.exceptions import HTTPException

        def _no_result_found_handler(e: NoResultFound) -> HTTPException:
            return HTTPException(404, str(e))

        # NoResultFound exceptions should be ignored since they will automatically
        # be converted to 404 responses.
        handler.ignore(NoResultFound)
        handler.prepare_using(NoResultFound, _no_result_found_handler)



import inspect

from typing import Any
from typing import TypeVar
from typing import cast
from typing import overload

import sqlalchemy.engine.url as _url

from sqlalchemy import URL
from sqlalchemy import event
from sqlalchemy import exc
from sqlalchemy import util
from sqlalchemy.engine import Engine
from sqlalchemy.engine import base
from sqlalchemy.engine.create import _pool_translate_kwargs
from sqlalchemy.engine.interfaces import DBAPIConnection
from sqlalchemy.engine.mock import create_mock_engine
from sqlalchemy.pool import ConnectionPoolEntry
from sqlalchemy.pool import _AdhocProxiedConnection
from sqlalchemy.sql import compiler


_T = TypeVar("_T", bound=Engine)


@overload
def create_engine(url: str | URL, engine_class: type[_T], **kwargs: Any) -> _T: ...


@overload
def create_engine(
    url: str | URL, engine_class: None = None, **kwargs: Any
) -> base.Engine: ...


def create_engine(
    url: str | URL, engine_class: type[_T] | None = None, **kwargs: Any
) -> _T | base.Engine:
    """
    This is a direct copy of the create_engine function from SQLAlchemy with added
    support for specifying the engine class. This is mainly useful to help with typing
    when having a custom connection class that extends the base Connection class.
    """

    if "strategy" in kwargs:
        strat = kwargs.pop("strategy")
        if strat == "mock":
            # this case is deprecated
            return create_mock_engine(url, **kwargs)  # type: ignore  # noqa: PGH003
        else:
            raise exc.ArgumentError(f"unknown strategy: {strat!r}")

    kwargs.pop("empty_in_strategy", None)

    # create url.URL object
    u = _url.make_url(url)

    u, plugins, kwargs = u._instantiate_plugins(kwargs)

    entrypoint = u._get_entrypoint()
    _is_async = kwargs.pop("_is_async", False)
    if _is_async:
        dialect_cls = entrypoint.get_async_dialect_cls(u)
    else:
        dialect_cls = entrypoint.get_dialect_cls(u)

    if kwargs.pop("_coerce_config", False):

        def pop_kwarg(key: str, default: Any | None = None) -> Any:
            value = kwargs.pop(key, default)
            if key in dialect_cls.engine_config_types:
                value = dialect_cls.engine_config_types[key](value)
            return value

    else:
        pop_kwarg = kwargs.pop  # type: ignore  # noqa: PGH003

    dialect_args = {}
    # consume dialect arguments from kwargs
    for k in util.get_cls_kwargs(dialect_cls):
        if k in kwargs:
            dialect_args[k] = pop_kwarg(k)

    dbapi = kwargs.pop("module", None)
    if dbapi is None:
        dbapi_args = {}

        if "import_dbapi" in dialect_cls.__dict__:
            dbapi_meth = dialect_cls.import_dbapi

        elif hasattr(dialect_cls, "dbapi") and inspect.ismethod(dialect_cls.dbapi):
            util.warn_deprecated(
                "The dbapi() classmethod on dialect classes has been "
                "renamed to import_dbapi().  Implement an import_dbapi() "
                f"classmethod directly on class {dialect_cls} to remove this "
                "warning; the old .dbapi() classmethod may be maintained for "
                "backwards compatibility.",
                "2.0",
            )
            dbapi_meth = dialect_cls.dbapi
        else:
            dbapi_meth = dialect_cls.import_dbapi

        for k in util.get_func_kwargs(dbapi_meth):
            if k in kwargs:
                dbapi_args[k] = pop_kwarg(k)
        dbapi = dbapi_meth(**dbapi_args)

    dialect_args["dbapi"] = dbapi

    dialect_args.setdefault("compiler_linting", compiler.NO_LINTING)
    enable_from_linting = kwargs.pop("enable_from_linting", True)
    if enable_from_linting:
        dialect_args["compiler_linting"] ^= compiler.COLLECT_CARTESIAN_PRODUCTS

    for plugin in plugins:
        plugin.handle_dialect_kwargs(dialect_cls, dialect_args)

    # create dialect
    dialect = dialect_cls(**dialect_args)

    # assemble connection arguments
    (cargs_tup, cparams) = dialect.create_connect_args(u)
    cparams.update(pop_kwarg("connect_args", {}))

    if "async_fallback" in cparams and util.asbool(cparams["async_fallback"]):
        util.warn_deprecated(
            "The async_fallback dialect argument is deprecated and will be "
            "removed in SQLAlchemy 2.1.",
            "2.0",
        )

    cargs = list(cargs_tup)  # allow mutability

    # look for existing pool or create
    pool = pop_kwarg("pool", None)
    if pool is None:

        def connect(
            connection_record: ConnectionPoolEntry | None = None,
        ) -> DBAPIConnection:
            if dialect._has_events:
                for fn in dialect.dispatch.do_connect:
                    connection = cast(
                        DBAPIConnection,
                        fn(dialect, connection_record, cargs, cparams),
                    )
                    if connection is not None:
                        return connection

            return dialect.connect(*cargs, **cparams)

        creator = pop_kwarg("creator", connect)

        poolclass = pop_kwarg("poolclass", None)
        if poolclass is None:
            poolclass = dialect.get_dialect_pool_class(u)
        pool_args = {"dialect": dialect}

        # consume pool arguments from kwargs, translating a few of
        # the arguments
        for k in util.get_cls_kwargs(poolclass):
            tk = _pool_translate_kwargs.get(k, k)
            if tk in kwargs:
                pool_args[k] = pop_kwarg(tk)

        for plugin in plugins:
            plugin.handle_pool_kwargs(poolclass, pool_args)

        pool = poolclass(creator, **pool_args)
    else:
        pool._dialect = dialect

    # create engine.
    if not pop_kwarg("future", True):
        raise exc.ArgumentError(
            "The 'future' parameter passed to "
            "create_engine() may only be set to True."
        )

    engineclass = engine_class or base.Engine

    engine_args = {}
    for k in util.get_cls_kwargs(engineclass):
        if k in kwargs:
            engine_args[k] = pop_kwarg(k)

    # internal flags used by the test suite for instrumenting / proxying
    # engines with mocks etc.
    _initialize = kwargs.pop("_initialize", True)

    # all kwargs should be consumed
    if kwargs:
        raise TypeError(
            "Invalid argument(s) {} sent to create_engine(), "
            "using configuration {}/{}/{}.  Please check that the "
            "keyword arguments are appropriate for this combination "
            "of components.".format(
                ",".join(f"'{k}'" for k in kwargs),
                dialect.__class__.__name__,
                pool.__class__.__name__,
                engineclass.__name__,
            )
        )

    engine = engineclass(pool, dialect, u, **engine_args)

    if _initialize:
        do_on_connect = dialect.on_connect_url(u)
        if do_on_connect:

            def on_connect(
                dbapi_connection: DBAPIConnection,
                connection_record: ConnectionPoolEntry,
            ) -> None:
                assert do_on_connect is not None
                do_on_connect(dbapi_connection)

            event.listen(pool, "connect", on_connect)

        builtin_on_connect = dialect._builtin_onconnect()
        if builtin_on_connect:
            event.listen(pool, "connect", builtin_on_connect)

        def first_connect(
            dbapi_connection: DBAPIConnection,
            connection_record: ConnectionPoolEntry,
        ) -> None:
            c = base.Connection(
                engine,
                connection=_AdhocProxiedConnection(dbapi_connection, connection_record),
                _has_events=False,
                # reconnecting will be a reentrant condition, so if the
                # connection goes away, Connection is then closed
                _allow_revalidate=False,
                # dont trigger the autobegin sequence
                # within the up front dialect checks
                _allow_autobegin=False,
            )
            c._execution_options = util.EMPTY_DICT

            try:
                dialect.initialize(c)
            finally:
                # note that "invalidated" and "closed" are mutually
                # exclusive in 1.4 Connection.
                if not c.invalidated and not c.closed:
                    # transaction is rolled back otherwise, tested by
                    # test/dialect/postgresql/test_dialect.py
                    # ::MiscBackendTest::test_initial_transaction_state
                    dialect.do_rollback(c.connection)

        # previously, the "first_connect" event was used here, which was then
        # scaled back if the "on_connect" handler were present.  now,
        # since "on_connect" is virtually always present, just use
        # "connect" event with once_unless_exception in all cases so that
        # the connection event flow is consistent in all cases.
        event.listen(pool, "connect", first_connect, _once_unless_exception=True)

    dialect_cls.engine_created(engine)
    if entrypoint is not dialect_cls:
        entrypoint.engine_created(engine)

    for plugin in plugins:
        plugin.engine_created(engine)

    return engine



# ruff: noqa: I002
import asyncio
import builtins
import inspect
import logging
import types

from collections.abc import Awaitable
from collections.abc import Callable
from inspect import isasyncgenfunction
from inspect import isgeneratorfunction
from typing import Annotated
from typing import Any
from typing import Self
from typing import TypeVar
from typing import get_args
from typing import get_origin
from typing import overload

from expanse.asynchronous.support._concurrency import run_in_threadpool
from expanse.common.container.container import Container as BaseContainer
from expanse.common.container.exceptions import ContainerException
from expanse.common.container.exceptions import ResolutionException
from expanse.common.container.exceptions import UnboundAbstractException
from expanse.common.support._utils import string_to_class


T = TypeVar("T")

_builtins = [d for d in dir(builtins) if isinstance(getattr(builtins, d), type)]
_EMPTY = object()

logger = logging.getLogger(__name__)

_Callback = Callable[..., None] | Callable[..., Awaitable[None]]


class UnboundAbstractError(Exception): ...


class Container(BaseContainer):
    def __init__(self) -> None:
        super().__init__()

        self._terminating_callbacks: list[_Callback] = []

    async def build(
        self, concrete: type | str, args: tuple | None = None
    ) -> tuple[Any, _Callback | None]:
        if args is None:
            args = ()

        function: Callable[..., Any]
        is_class: bool = False
        if isinstance(concrete, types.FunctionType):
            if concrete.__name__ == "<lambda>":
                return concrete(self, *args), None

            if concrete.__name__ == "_container_concrete":
                return await concrete(self), None

            function = concrete
        elif isinstance(concrete, types.MethodType):
            function = concrete
        else:
            if isinstance(concrete, str):
                concrete = string_to_class(concrete)

            if isinstance(concrete, types.MethodType):
                function = concrete
            else:
                # What we are trying to build is a class,
                # so we need to resolve the parameters of the __init__ method.
                function = concrete.__init__  # type: ignore[misc]
                is_class = True

        (
            positional,
            keywords,
        ) = await self._resolve_callable_dependencies(function, *args)

        if isasyncgenfunction(concrete):
            generator = concrete(*positional, **keywords)

            async def terminating_callback() -> None:
                await anext(generator, None)

            return await anext(generator), terminating_callback

        if isgeneratorfunction(concrete):
            generator = concrete(*positional, **keywords)

            def sync_terminating_callback() -> None:
                next(generator, None)

            return next(generator), sync_terminating_callback

        # If we are trying to build a class, we want to instantiate it directly
        # without executing it in a worker thread since it has an impact on performance.
        if is_class:
            return concrete(*positional, **keywords), None

        if not asyncio.iscoroutinefunction(concrete):
            return await run_in_threadpool(concrete, *positional, **keywords), None

        return await concrete(*positional, **keywords), None

    @overload
    async def get(self, abstract: type[T]) -> T: ...

    @overload
    async def get(self, abstract: str) -> Any: ...

    async def get(self, abstract: str | type[T]) -> Any | T:
        return await self._resolve(abstract)

    async def call(
        self, callable: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> Any:
        if (
            isinstance(callable, types.FunctionType)
            and "." in callable.__qualname__
            and not inspect.ismethod(callable)
            and "<locals>" not in callable.__qualname__
        ):
            # We have an instance method, so we will retrieve the corresponding class,
            # resolve it and call the method.
            class_name, func_name = callable.__qualname__.rsplit(".", maxsplit=1)
            class_: type = callable.__globals__[class_name]

            instance: Any = await self.get(class_)

            callable = getattr(instance, func_name)

        (
            positional,
            keywords,
        ) = await self._resolve_callable_dependencies(callable, *args, **kwargs)

        if asyncio.iscoroutinefunction(callable):
            return await callable(*positional, **keywords)

        return await run_in_threadpool(callable, *positional, **keywords)

    def terminating(self, callback: _Callback, scoped: bool = False) -> None:
        if scoped:
            self._scoped["terminating_callbacks"].append(callback)
        else:
            self._terminating_callbacks.append(callback)

    async def terminate(self) -> None:
        for callback in self._terminating_callbacks:
            await self.call(callback)

    def create_scoped_container(self) -> "ScopedContainer":
        container = ScopedContainer(self)

        return container

    async def on_resolved(
        self,
        abstract: str | type,
        callback: _Callback,
    ) -> None:
        if not self._is_scoped(abstract) and self.resolved(abstract):
            if asyncio.iscoroutinefunction(callback):
                await callback(await self.get(abstract))
            else:
                callback(await self.get(abstract))

        self.after_resolving(abstract, callback)

    def _concrete_closure(
        self, abstract: str | type, concrete: Any
    ) -> Callable[[Self], Awaitable[Any]]:
        original_concrete = concrete

        async def closure(container: Container) -> Any:
            if abstract == original_concrete:
                obj, _ = await container.build(original_concrete)

                return obj

            return await container._resolve(original_concrete)

        return closure

    @overload
    async def _resolve(self, abstract: type[T]) -> T: ...

    @overload
    async def _resolve(self, abstract: str) -> Any: ...

    async def _resolve(self, abstract: str | type[T]) -> Any | T:
        abstract = self._get_alias(abstract)

        if abstract in self._instances:
            return self._instances[abstract]

        metadata: tuple = ()
        actual_abstract: str | type = abstract
        origin = get_origin(abstract)
        if origin is Annotated:
            actual_abstract, *metadata = get_args(abstract)  # type: ignore[assignment]

        if actual_abstract in self._bindings:
            concrete = self._bindings[actual_abstract]["concrete"]
        elif isinstance(abstract, str):
            # Unbound strings cannot be resolved
            raise UnboundAbstractException(
                f"Unbound abstract [{abstract}] cannot be resolved"
            )
        else:
            concrete = abstract

        terminating_callback: _Callback | None = None
        if self._can_build(actual_abstract, concrete):
            try:
                obj, terminating_callback = await self.build(concrete, metadata)
            except Exception as e:
                raise ContainerException(
                    f'Unable to build the "{abstract}" dependency'
                ) from e
        else:
            obj = await self.get(concrete)

        if self._is_cached(actual_abstract):
            self._instances[abstract] = obj

        self._mark_as_resolved(actual_abstract)
        if terminating_callback is not None:
            self.terminating(
                terminating_callback, scoped=self._is_scoped(actual_abstract)
            )

        await self._execute_after_resolving_callbacks(abstract, obj)

        return obj

    async def _resolve_callable_dependencies(
        self, callable: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> tuple[list[Any], dict[str, Any]]:
        positional: list[Any] = []
        keywords: dict[str, Any] = {}
        arguments = list(args)
        _globals = getattr(callable, "__globals__", None)

        for name, parameter in inspect.signature(callable).parameters.items():
            klass = self._get_class(parameter)

            if name == "self":
                continue

            if klass is None:
                await self._resolve_primitive(
                    parameter, arguments, kwargs, positional, keywords
                )
            else:
                try:
                    await self._resolve_class(
                        parameter,
                        arguments,
                        kwargs,
                        positional,
                        keywords,
                        _globals=_globals,
                    )
                except ContainerException as e:
                    raise ResolutionException(
                        f'Unable to resolve dependency with name "{parameter.name}" '
                        f'(type: {klass.__module__ + "." + klass.__qualname__}) '
                        f'in "{callable.__module__ + "." + callable.__qualname__}"'
                    ) from e

        return positional, keywords

    async def _resolve_primitive(
        self,
        parameter: inspect.Parameter,
        args: list[Any],
        kwargs: dict[str, Any],
        positional: list[Any],
        keywords: dict[str, Any],
    ) -> None:
        match parameter.kind:
            case parameter.POSITIONAL_ONLY:
                if not args:
                    raise ResolutionException(
                        f'Unable to resolve dependency with name "{parameter.name}"'
                    )

                positional.append(args.pop(0))

                return

            case parameter.POSITIONAL_OR_KEYWORD:
                # Check in keyword arguments first
                if parameter.name in kwargs:
                    keywords[parameter.name] = kwargs.pop(parameter.name)

                elif args:
                    positional.append(args.pop(0))

                elif parameter.default is not parameter.empty:
                    keywords[parameter.name] = parameter.default

                return

            case parameter.KEYWORD_ONLY:
                if parameter.name in kwargs:
                    keywords[parameter.name] = kwargs.pop(
                        parameter.name,
                    )
                elif parameter.default is not parameter.empty:
                    keywords[parameter.name] = parameter.default

                return

            case parameter.VAR_KEYWORD:
                keywords.update(kwargs.copy())

                kwargs.clear()

                return

            case parameter.VAR_POSITIONAL:
                positional.extend(args.copy())

                args.clear()

                return

            case _:
                pass

        raise ResolutionException(
            f'Unable to resolve dependency with name "{parameter.name}"'
        )

    async def _resolve_class(
        self,
        parameter: inspect.Parameter,
        args: list[Any],
        kwargs: dict[str, Any],
        positional: list[Any],
        keywords: dict[str, Any],
        *,
        _globals: dict[str, Any] | None = None,
    ) -> Any:
        match parameter.kind:
            case parameter.POSITIONAL_ONLY:
                klass = self._get_class(parameter, _globals=_globals)

                assert klass is not None

                if self.has(self._get_alias(klass)):
                    result = await self.get(self._get_alias(klass))
                else:
                    try:
                        result = await self.get(self._get_alias(klass))
                    except Exception as e:
                        if not args:
                            raise e

                        arg = args[0]

                        if not isinstance(arg, klass):
                            raise e

                        result = args.pop(0)

                positional.append(result)
                return

            case parameter.POSITIONAL_OR_KEYWORD:
                # Check in keyword arguments first
                if parameter.name in kwargs:
                    keywords[parameter.name] = kwargs.pop(parameter.name)
                else:
                    klass = self._get_class(parameter, _globals=_globals)

                    assert klass is not None

                    if self.has(self._get_alias(klass)):
                        result = await self.get(self._get_alias(klass))
                    else:
                        try:
                            result = await self.get(self._get_alias(klass))
                        except Exception as e:
                            if not args:
                                raise e

                            arg = args[0]

                            if not isinstance(arg, klass):
                                raise e

                            result = args.pop(0)

                    positional.append(result)

                return

            case parameter.KEYWORD_ONLY:
                if parameter.name in kwargs:
                    keywords[parameter.name] = kwargs.pop(
                        parameter.name,
                    )
                else:
                    klass = self._get_class(parameter, _globals=_globals)

                    assert klass is not None

                    result = await self.get(self._get_alias(klass))

                    keywords[parameter.name] = result
                return

            case parameter.VAR_POSITIONAL:
                klass = self._get_class(parameter, _globals=_globals)

                assert klass is not None

                result = await self.get(self._get_alias(klass))

                result = [result] if not isinstance(result, tuple) else result

                positional.extend(result)
                return

            case _:
                pass

        raise ResolutionException(
            f'Unable to resolve dependency with name "{parameter.name}"'
        )

    async def _execute_after_resolving_callbacks(
        self, abstract: str | type, instance: Any
    ) -> None:
        callbacks: list[Callable[..., Awaitable[None]]] = []
        abstract = self._get_alias(abstract)

        actual_abstract: str | type = abstract
        origin = get_origin(abstract)
        if origin is Annotated:
            actual_abstract, *_ = get_args(abstract)

        if abstract in self._after_resolving_callbacks:
            callbacks += self._after_resolving_callbacks[abstract]
        elif actual_abstract in self._after_resolving_callbacks:
            callbacks += self._after_resolving_callbacks[actual_abstract]

        for callback in callbacks:
            if asyncio.iscoroutinefunction(callback):
                await callback(instance)
            else:
                callback(instance)

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ) -> None:
        await self.terminate()


class ScopedContainer(Container):
    def __init__(self, base_container: Container):
        super().__init__()

        self._base_container = base_container

        # Bind scoped bindings from the base container
        self._bindings.update(
            {k: {**v} for k, v in self._base_container._scoped["bindings"].items()}
        )

        # Setup terminating callbacks
        self._terminating_callbacks = [
            *self._base_container._scoped["terminating_callbacks"]
        ]

        # Setup resolving callbacks
        self._after_resolving_callbacks = {
            **self._base_container._scoped["after_resolving_callbacks"]
        }

        # Copy instances from the base container
        self._instances.update(self._base_container._instances)

        self.instance(Container, self)

    def bound(self, abstract: str | type) -> bool:
        return self._base_container.bound(abstract) or super().bound(abstract)

    def _directly_bound(self, abstract: str | type) -> bool:
        return super().bound(abstract)

    async def _resolve(self, abstract: str | type[T]) -> Any | T:
        actual_abstract: str | type[T] = abstract
        origin = get_origin(abstract)
        if origin is Annotated:
            actual_abstract, *_ = get_args(abstract)

        # If the abstract is neither bound in the container nor in its base container,
        # we will resolve it from the scoped container.
        if not self.bound(abstract):
            return await super()._resolve(abstract)

        if not self._directly_bound(actual_abstract):
            return await self._base_container._resolve(abstract)

        return await super()._resolve(abstract)



import logging

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import Protocol

from alembic.runtime.migration import RevisionStep


if TYPE_CHECKING:
    from logging import LogRecord

    from cleo.io.io import IO


class Formatter(Protocol):
    def format(self, msg: "LogRecord") -> str | None: ...


class MigrationFormatter:
    def format(self, record: "LogRecord") -> str | None:
        msg = record.msg
        if msg.startswith(("Context impl ", "Will assume ")):
            return None
        elif msg.startswith("Running "):
            assert isinstance(record.args, tuple)
            assert isinstance(record.args[0], RevisionStep)
            # First arg should be the revision step
            step: RevisionStep = record.args[0]
            msg_parts = [
                "  -",
                "<comment>Applying</comment>"
                if step.is_upgrade
                else "<fg=yellow>Rolling back</>",
                "migration",
                f"<info>{step.revision.revision}</info>",
                f"<options=dark>({step.revision.doc})</>",
            ]
            record.args = record.args[1:]
            return " ".join(msg_parts)

        return msg


class IOFormatter(logging.Formatter):
    _colors: ClassVar[dict[str, str]] = {
        "error": "fg=red",
        "warning": "fg=yellow",
        "debug": "debug",
        "info": "fg=blue",
    }

    def __init__(self, formatters: dict[str, Formatter]) -> None:
        super().__init__()

        self._formatters = formatters

    def format(self, record: "LogRecord") -> str:
        if not record.exc_info:
            level = record.levelname.lower()
            msg = record.msg

            if record.name in self._formatters:
                msg = self._formatters[record.name].format(record)
                if msg is None:
                    return ""
            elif level in self._colors:
                msg = f"<{self._colors[level]}>{msg}</>"

            record.msg = msg

        formatted = super().format(record)

        return formatted


class IOHandler(logging.Handler):
    def __init__(self, io: "IO") -> None:
        self._io = io

        super().__init__()

    def emit(self, record: "LogRecord") -> None:
        try:
            msg = self.format(record)
            level = record.levelname.lower()
            err = level in ("warning", "error", "exception", "critical")
            if err:
                self._io.write_error_line(msg)
            elif msg:
                self._io.write_line(msg)
        except Exception:
            self.handleError(record)


def configure_alembic_loggers(io: "IO", disable: bool = False) -> None:
    """
    Configure Alembic loggers so that the output is nicer than the default one.
    """
    logger = logging.getLogger("alembic.runtime.migration")

    if disable:
        for handler in logger.handlers:
            logger.removeHandler(handler)
    else:
        level = logging.INFO
        if io.is_debug():
            level = logging.DEBUG

        handler = IOHandler(io)
        handler.setFormatter(
            IOFormatter({"alembic.runtime.migration": MigrationFormatter()})
        )
        handler.setLevel(level)
        logger.addHandler(handler)
        logger.setLevel(level)



from pathlib import Path

import pytest

from expanse.core.application import Application
from expanse.database.console.commands.db_rollback import (
    DbRollbackCommand,  # noqa: F401
)
from expanse.testing.command_tester import CommandTester


@pytest.fixture(autouse=True)
def migrate(command_tester: CommandTester, app: Application) -> None:
    app.config["paths"]["database"] = (
        Path(__file__).parent.joinpath("fixtures").relative_to(app.base_path)
    )
    app.config["database"]["default"] = "test"

    command = command_tester.command("db migrate")

    command.run()


def test_rollback(command_tester: CommandTester, app: Application) -> None:
    command = command_tester.command("db rollback")

    return_code = command.run()
    assert return_code == 0

    expected = """
  - Rolling back migration 1234567891 (Foo Migration)
"""

    assert command.output.fetch() == expected


def test_rollback_with_step(command_tester: CommandTester, app: Application) -> None:
    command = command_tester.command("db rollback")

    return_code = command.run("--step 2")
    assert return_code == 0

    expected = """
  - Rolling back migration 1234567891 (Foo Migration)
  - Rolling back migration 1234567890 (Auto Migration)
"""

    assert command.output.fetch() == expected



from pathlib import Path

from pydantic import HttpUrl

from expanse.asynchronous.http.helpers import view
from expanse.asynchronous.http.response import Response
from expanse.asynchronous.routing.route import Route
from expanse.asynchronous.routing.router import Router
from expanse.asynchronous.static.static_service_provider import StaticServiceProvider
from expanse.asynchronous.testing.client import TestClient
from expanse.asynchronous.view.view_service_provider import ViewServiceProvider


FIXTURES_DIR = Path(__file__).parent / "fixtures"


async def test_endpoint_returns_static_content_successfully(client: TestClient) -> None:
    provider = StaticServiceProvider(client.app.container)

    client.app.config["static.paths"] = [FIXTURES_DIR]
    await client.app.register(provider)
    await provider.boot()

    response = client.get("/static/foo.txt")
    assert response.status_code == 200
    assert response.text.strip() == "Foo"


async def test_endpoint_is_not_set_up_if_not_debug_mode(client: TestClient) -> None:
    provider = StaticServiceProvider(client.app.container)

    client.app.config["app.debug"] = False
    client.app.config["static.paths"] = [FIXTURES_DIR]
    await client.app.register(provider)
    await provider.boot()

    response = client.get("/static/foo.txt")
    assert response.status_code == 404


async def test_view_static_function_is_registered_successfully(
    client: TestClient,
) -> None:
    provider = StaticServiceProvider(client.app.container)

    client.app.config["static.paths"] = [FIXTURES_DIR]
    client.app.config["view.paths"] = [FIXTURES_DIR]
    await client.app.register(ViewServiceProvider(client.app.container))
    await client.app.register(provider)
    await provider.boot()

    async def foo() -> Response:
        return await view("foo")

    (await client.app.container.get(Router)).add_route(Route.get("/foo", foo))

    response = client.get("/foo")
    assert response.status_code == 200
    assert "/static/foo.txt" in response.text


async def test_static_url_includes_base_url(client: TestClient) -> None:
    provider = StaticServiceProvider(client.app.container)

    client.app.config["static.paths"] = [FIXTURES_DIR]
    client.app.config["static.prefix"] = "/assets/"
    client.app.config["static.url"] = HttpUrl("https://assets.example.com/")
    client.app.config["view.paths"] = [FIXTURES_DIR]
    await client.app.register(ViewServiceProvider(client.app.container))
    await client.app.register(provider)
    await provider.boot()

    async def foo() -> Response:
        return await view("foo")

    (await client.app.container.get(Router)).add_route(Route.get("/foo", foo))

    response = client.get("/foo")
    assert response.status_code == 200
    assert "https://assets.example.com/assets/foo.txt" in response.text


async def test_static_url_with_empty_prefix(client: TestClient) -> None:
    provider = StaticServiceProvider(client.app.container)

    client.app.config["app.debug"] = False
    client.app.config["static.paths"] = [FIXTURES_DIR]
    client.app.config["static.prefix"] = ""
    client.app.config["static.url"] = HttpUrl("https://assets.example.com/")
    client.app.config["view.paths"] = [FIXTURES_DIR]
    await client.app.register(ViewServiceProvider(client.app.container))
    await client.app.register(provider)
    await provider.boot()

    async def foo() -> Response:
        return await view("foo")

    (await client.app.container.get(Router)).add_route(Route.get("/foo", foo))

    response = client.get("/foo")
    assert response.status_code == 200
    assert "https://assets.example.com/foo.txt" in response.text



from pathlib import Path

from expanse.asynchronous.http.request import Request
from expanse.asynchronous.http.response import Response
from expanse.asynchronous.routing.redirect import Redirect
from expanse.asynchronous.routing.responder import Responder
from expanse.asynchronous.routing.router import Router
from expanse.asynchronous.view.view_factory import ViewFactory
from expanse.asynchronous.view.view_finder import ViewFinder


def test_redirect(router: Router) -> None:
    redirect = Redirect(router, Request.create("http://example.com"))

    responder = Responder(ViewFactory(ViewFinder([])), redirect)

    assert responder.redirect() == redirect


async def test_text(router: Router) -> None:
    redirect = Redirect(router, Request.create("http://example.com"))

    responder = Responder(
        ViewFactory(ViewFinder([])),
        redirect,
    )

    response = await responder.text("view", headers={"X-Header": "foo"})

    assert isinstance(response, Response)
    assert response.status_code == 200
    assert response.content_type == "text/plain"
    assert response.charset == "utf-8"
    assert response.headers["X-Header"] == "foo"


async def test_html(router: Router) -> None:
    redirect = Redirect(router, Request.create("http://example.com"))

    responder = Responder(
        ViewFactory(ViewFinder([])),
        redirect,
    )

    response = await responder.html("view", headers={"X-Header": "foo"})

    assert isinstance(response, Response)
    assert response.status_code == 200
    assert response.content_type == "text/html"
    assert response.charset == "utf-8"
    assert response.headers["X-Header"] == "foo"


async def test_json(router: Router) -> None:
    redirect = Redirect(router, Request.create("http://example.com"))

    responder = Responder(
        ViewFactory(ViewFinder([])),
        redirect,
    )

    response = await responder.json({"foo": "bar"}, headers={"X-Header": "foo"})

    assert isinstance(response, Response)
    assert response.status_code == 200
    assert response.content_type == "application/json"
    assert response.charset == "utf-8"
    assert response.headers["X-Header"] == "foo"


async def test_file(router: Router) -> None:
    redirect = Redirect(router, Request.create("http://example.com"))

    responder = Responder(
        ViewFactory(ViewFinder([])),
        redirect,
    )

    response = await responder.file(
        Path(__file__).parent.joinpath("fixtures/files/text.txt"),
        headers={"X-Header": "foo"},
    )

    assert isinstance(response, Response)
    assert response.status_code == 200
    assert response.content_type == "text/plain"
    assert response.charset is None
    assert response.headers["X-Header"] == "foo"


async def test_view(router: Router) -> None:
    redirect = Redirect(router, Request.create("http://example.com"))

    responder = Responder(
        ViewFactory(ViewFinder([Path(__file__).parent.joinpath("fixtures/views")])),
        redirect,
    )

    response = await responder.view("view", headers={"X-Header": "foo"})

    assert isinstance(response, Response)
    assert response.status_code == 200
    assert response.content_type == "text/html"
    assert response.charset == "utf-8"
    assert response.headers["X-Header"] == "foo"



import base64
import re

from typing import ClassVar

from cleo.helpers import option
from cleo.io.inputs.option import Option

from expanse.common.encryption.encryptor import Cipher
from expanse.common.encryption.encryptor import Encryptor
from expanse.console.commands.command import Command
from expanse.core.application import Application


class EncryptionKeyGenerateCommand(Command):
    name: str = "encryption key generate"
    description: str = "Generate a new encryption key."

    options: ClassVar[list[Option]] = [
        option(
            "show", None, "Only display the encryption key without modifying files."
        ),
    ]

    def handle(self, app: Application) -> int:
        raw_key = Encryptor.generate_random_key(
            Cipher(app.config.get("encryption.cipher"))
        )
        key = f"base64:{base64.urlsafe_b64encode(raw_key).decode()}"

        if self.option("show"):
            self.line("")
            self.line(f"<c1>{key}</c1>")

            return 0

        if not self._replace_key(app, key):
            return 1

        app.config["app.secret_key"] = key

        self.line("Application secret key generated successfully.", style="success")

        return 0

    def _replace_key(self, app: Application, key: str) -> bool:
        existing_key: str | None = app.config.get("app.secret_key")

        if existing_key and not self.confirm(
            "Do you want to replace the existing application secret key?"
        ):
            return False

        return self._write_key(app, key)

    def _write_key(self, app: Application, key: str) -> bool:
        env_file = app.environment_path / app.environment_file
        if not env_file.exists():
            self.line_error("The .env file does not exist.", style="error")

            return False

        content = env_file.read_text()

        escaped = re.escape(f"={app.config.get('app.secret_key', '')}")

        new_content = re.sub(
            rf"^APP_SECRET_KEY{escaped}",
            f"APP_SECRET_KEY={key}",
            content,
            flags=re.MULTILINE,
        )

        if new_content == content:
            self.line_error(
                "The application key could not be set. "
                "Please check your .env file and that a APP_SECRET_KEY is present.",
                style="error",
            )

            return False

        env_file.write_text(new_content)

        return True



import re

from typing import Any
from urllib.parse import urlencode

from expanse.common.http.url import URL
from expanse.http.request import Request
from expanse.routing.router import Router


class URLGenerator:
    def __init__(self, router: Router, request: Request) -> None:
        self._router: Router = router
        self._request: Request = request

    def to(
        self,
        path: str,
        secure: bool | None = None,
    ) -> str:
        # If the URL is already valid, there is nothing to do, and we return it directly.
        if self._is_valid_url(path):
            return path

        url = self._request.url

        if secure is not None:
            url = url.replace(scheme="https" if secure else "http")

        return str(url.replace(path=path))

    def to_route(
        self,
        name: str,
        parameters: dict[str, Any] | None = None,
        absolute: bool = False,
    ) -> str:
        # If the URL is already valid, there is nothing to do, and we return it directly.
        url = self._router.route(name, parameters)

        if absolute:
            return str(
                URL(url).replace(
                    scheme=self._request.url.scheme, netloc=self._request.url.netloc
                )
            )

        return str(url)

    def _format_parameters(self, parameters: dict[str, Any] | None) -> str:
        if parameters is None:
            return ""

        return urlencode(parameters)

    def _is_valid_url(self, path: str) -> bool:
        return re.match(r"^(#|//|https?://|(mailto|tel|sms):)", path) is not None



from expanse.asynchronous.http.helpers import json
from expanse.asynchronous.http.response import Response
from expanse.asynchronous.routing.router import Router
from expanse.asynchronous.testing.client import TestClient
from expanse.common.http.json import JSON
from tests.asynchronous.integration.http.fixtures.request.models import FooModel


async def create_foo_validated(form: JSON[FooModel]) -> Response:
    return await json({"bar": form.bar})


def test_simple_json_data_are_converted_if_validation_model(
    router: Router, client: TestClient
) -> None:
    router.post("/", create_foo_validated)

    response = client.post("/", json={"bar": "42"})

    assert response.json() == {"bar": 42}


def test_validation_errors_are_correctly_reported(
    router: Router, client: TestClient
) -> None:
    router.post("/", create_foo_validated)

    response = client.post("/", json={"bar": "foo"})

    assert response.status_code == 422
    assert response.json() == {
        "code": "validation_error",
        "detail": [
            {
                "loc": ["bar"],
                "message": (
                    "Input should be a valid integer, "
                    "unable to parse string as an integer"
                ),
                "type": "int_parsing",
            }
        ],
    }



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING
from typing import Any
from typing import Self

from baize.wsgi.requests import Request as BaseRequest

from expanse.common.http.accept_header import AcceptHeader
from expanse.common.http.url import URL


if TYPE_CHECKING:
    from expanse.routing.route import Route
    from expanse.types import Environ
    from expanse.types import StartResponse


class Request(BaseRequest):
    def __init__(self, environ: Environ, start_response: StartResponse | None = None):
        super().__init__(environ, start_response)  # type: ignore[arg-type]

        self._route: Route | None = None

        self._acceptable_content_types: list[str] | None = None

    @cached_property
    def url(self) -> URL:  # type: ignore[override]
        return URL(environ=self._environ)

    @cached_property
    def host(self) -> str:
        client = self.client

        if not client:
            return ""

        return client.host or ""

    @cached_property
    def acceptable_content_types(self) -> list[str]:
        return [
            item.value
            for item in AcceptHeader.from_string(self.headers.get("Accept", "")).all()
        ]

    def accepts_any_content_type(self) -> bool:
        """
        Determine if the current request accepts any content type.
        """
        acceptable = self.acceptable_content_types

        return len(acceptable) == 0 or acceptable[0] in ("*/*", "*")

    def is_json(self) -> bool:
        """
        Determine whether the request is sending JSON or not.
        """
        return "/json" in self.content_type.type or "+json" in self.content_type.type

    def wants_json(self) -> bool:
        """
        Determine whether the request is asking for JSON or not.
        """
        acceptable = self.acceptable_content_types

        return len(acceptable) > 0 and (
            "/json" in acceptable[0] or "+json" in acceptable[0]
        )

    def expects_json(self) -> bool:
        """
        Determine if the current request probably expects a JSON response.
        """
        return (
            self.is_ajax()
            and not self.is_pjax()
            and self.accepts_any_content_type()
            or self.wants_json()
        )

    def is_xml_http_request(self) -> bool:
        return self.headers.get("X-Requested-With") == "XMLHttpRequest"

    def is_ajax(self) -> bool:
        return self.is_xml_http_request()

    def is_pjax(self) -> bool:
        return self.headers.get("X-PJAX") == "true"

    @property
    def route(self) -> Route | None:
        return self._route

    def set_route(self, route: Route) -> Self:
        self._route = route

        return self

    @classmethod
    def create(
        cls, raw_url: str, method: str = "GET", environ: dict[str, Any] | None = None
    ) -> Request:
        base_environ = {
            "wsgi.url_scheme": "http",
            "SERVER_NAME": "localhost",
            "SERVER_PORT": 80,
            "HTTP_HOST": "localhost",
            "HTTP_USER_AGENT": "Expanse",
            "HTTP_ACCEPT": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
            "HTTP_ACCEPT_LANGUAGE": "en-us,en;q=0.5",
            "REMOTE_ADDR": "127.0.0.1",
            "SCRIPT_NAME": "",
            "SCRIPT_FILENAME": "",
            "SERVER_PROTOCOL": "HTTP/1.1",
            "PATH_INFO": "",
            "REQUEST_METHOD": method.upper(),
            **(environ or {}),
        }

        url = URL(raw_url)

        if url.hostname is not None:
            base_environ["SERVER_NAME"] = url.hostname
            base_environ["HTTP_HOST"] = url.hostname

        if url.scheme:
            base_environ["wsgi.url_scheme"] = url.scheme

            if url.scheme == "https":
                base_environ["HTTPS"] = "on"
                base_environ["SERVER_PORT"] = 443
            else:
                base_environ.pop("HTTPS", None)
                base_environ["SERVER_PORT"] = 80

        if url.port is not None:
            base_environ["SERVER_PORT"] = url.port
            base_environ["HTTP_HOST"] += f":{url.port}"

        path = url.path
        if not path:
            path = "/"

        base_environ["PATH_INFO"] = path

        query_string = ""
        if url.query:
            query_string = url.query

        base_environ["REQUEST_URI"] = path + (
            "?" + query_string if query_string else ""
        )
        base_environ["QUERY_STRING"] = query_string

        return cls(environ=base_environ)


__all__ = ["Request"]



from typing import cast

from sqlalchemy.engine import Engine as BaseEngine

from expanse.database.connection import Connection


class Engine(BaseEngine):
    _connection_cls = Connection

    def connect(self) -> Connection:
        return cast(Connection, super().connect())



from pathlib import Path

from pydantic import Field
from pydantic_settings import BaseSettings
from pydantic_settings import SettingsConfigDict


class Config(BaseSettings):
    paths: list[Path] = Field(default=[Path("resources/views")])
    model_config = SettingsConfigDict(env_prefix="view_")



import contextlib
import io

from collections.abc import Callable
from collections.abc import Generator
from contextlib import AbstractContextManager
from types import GeneratorType
from typing import Any
from typing import Literal
from typing import TypedDict
from urllib.parse import unquote

import anyio
import anyio.abc
import anyio.from_thread
import httpx

from expanse.asynchronous.contracts.debug.exception_handler import ExceptionHandler
from expanse.asynchronous.core.application import Application
from expanse.asynchronous.types import Message
from expanse.common.testing.client import TestClient as BaseTestClient


_PortalFactoryType = Callable[[], AbstractContextManager[anyio.abc.BlockingPortal]]


class _AsyncBackend(TypedDict):
    backend: str
    backend_options: dict[str, Any]


class _TestClientTransport(httpx.BaseTransport):
    def __init__(
        self,
        app: Application,
        portal_factory: _PortalFactoryType,
        root_path: str = "",
        raise_server_exceptions: bool = False,
    ) -> None:
        self.app = app
        self.portal_factory = portal_factory
        self.root_path = root_path
        self.raise_server_exceptions = raise_server_exceptions

    def handle_request(self, request: httpx.Request) -> httpx.Response:
        scheme = request.url.scheme
        netloc = request.url.netloc.decode(encoding="ascii")
        path = request.url.path
        raw_path = request.url.raw_path
        query = request.url.query.decode(encoding="ascii")

        default_port = {"http": 80, "ws": 80, "https": 443, "wss": 443}[scheme]

        if ":" in netloc:
            host, port_string = netloc.split(":", 1)
            port = int(port_string)
        else:
            host = netloc
            port = default_port

        # Include the 'host' header.
        if "host" in request.headers:
            headers: list[tuple[bytes, bytes]] = []
        elif port == default_port:  # pragma: no cover
            headers = [(b"host", host.encode())]
        else:  # pragma: no cover
            headers = [(b"host", (f"{host}:{port}").encode())]

        # Include other request headers.
        headers += [
            (key.lower().encode(), value.encode())
            for key, value in request.headers.multi_items()
        ]

        scope: dict[str, Any]

        scope = {
            "type": "http",
            "http_version": "1.1",
            "method": request.method,
            "path": unquote(path),
            "raw_path": raw_path,
            "root_path": self.root_path,
            "scheme": scheme,
            "query_string": query.encode(),
            "headers": headers,
            "client": None,
            "server": [host, port],
            "extensions": {"http.response.debug": {}},
        }

        request_complete = False
        response_started = False
        response_complete: anyio.Event
        raw_kwargs: dict[str, Any] = {"stream": io.BytesIO()}
        template = None
        context = None

        async def receive() -> Message:
            nonlocal request_complete

            if request_complete:
                if not response_complete.is_set():
                    await response_complete.wait()
                return {"type": "http.disconnect"}

            body = request.read()
            if isinstance(body, str):
                body_bytes: bytes = body.encode("utf-8")  # pragma: no cover
            elif body is None:
                body_bytes = b""  # pragma: no cover
            elif isinstance(body, GeneratorType):
                try:  # pragma: no cover
                    chunk = body.send(None)
                    if isinstance(chunk, str):
                        chunk = chunk.encode("utf-8")
                    return {"type": "http.request", "body": chunk, "more_body": True}
                except StopIteration:  # pragma: no cover
                    request_complete = True
                    return {"type": "http.request", "body": b""}
            else:
                body_bytes = body

            request_complete = True
            return {"type": "http.request", "body": body_bytes}

        async def send(message: Message) -> None:
            nonlocal raw_kwargs, response_started, template, context

            if message["type"] == "http.response.start":
                assert (
                    not response_started
                ), 'Received multiple "http.response.start" messages.'
                raw_kwargs["status_code"] = message["status"]
                raw_kwargs["headers"] = [
                    (key.decode(), value.decode())
                    for key, value in message.get("headers", [])
                ]
                response_started = True
            elif message["type"] == "http.response.body":
                assert (
                    response_started
                ), 'Received "http.response.body" without "http.response.start".'
                assert (
                    not response_complete.is_set()
                ), 'Received "http.response.body" after response completed.'
                body = message.get("body", b"")
                more_body = message.get("more_body", False)
                if request.method != "HEAD":
                    raw_kwargs["stream"].write(body)
                if not more_body:
                    raw_kwargs["stream"].seek(0)
                    response_complete.set()
            elif message["type"] == "http.response.debug":
                template = message["info"]["template"]
                context = message["info"]["context"]

        with self.portal_factory() as portal:
            response_complete = portal.call(anyio.Event)
            handler: ExceptionHandler = portal.call(
                self.app.container.get, ExceptionHandler
            )
            with handler.raise_unhandled_exceptions(self.raise_server_exceptions):
                portal.call(self.app, scope, receive, send)

        if self.raise_server_exceptions:
            assert response_started, "TestClient did not receive any response."
        elif not response_started:
            raw_kwargs = {
                "status_code": 500,
                "headers": [],
                "stream": io.BytesIO(),
            }

        raw_kwargs["stream"] = httpx.ByteStream(raw_kwargs["stream"].read())

        response = httpx.Response(**raw_kwargs, request=request)
        if template is not None:
            response.template = template  # type: ignore[attr-defined]
            response.context = context  # type: ignore[attr-defined]
        return response


class TestClient(BaseTestClient[Application]):
    def __init__(
        self,
        app: Application,
        base_url: str = "http://testserver",
        backend: Literal["asyncio", "trio"] = "asyncio",
        backend_options: dict[str, Any] | None = None,
        cookies: httpx._client.CookieTypes = None,
        headers: dict[str, str] | None = None,
        raise_server_exceptions: bool = False,
    ) -> None:
        self._transport: _TestClientTransport | None = None
        self.app = app
        self.async_backend = _AsyncBackend(
            backend=backend, backend_options=backend_options or {}
        )
        self.raise_server_exceptions = raise_server_exceptions
        if headers is None:
            headers = {}

        headers.setdefault("user-agent", "testclient")
        super().__init__(
            app=self.app,
            base_url=base_url,
            headers=headers,
            follow_redirects=True,
            cookies=cookies,
        )

    @property
    def transport(self) -> _TestClientTransport:
        if self._transport is None:
            self._transport = _TestClientTransport(
                self.app,
                portal_factory=self._portal_factory,
                raise_server_exceptions=self.raise_server_exceptions,
            )

        return self._transport

    @contextlib.contextmanager
    def _portal_factory(
        self,
    ) -> Generator[anyio.abc.BlockingPortal, None, None]:
        with anyio.from_thread.start_blocking_portal(**self.async_backend) as portal:
            yield portal

    @contextlib.contextmanager
    def handle_exceptions(
        self, handle_exceptions: bool = True
    ) -> Generator[None, None, None]:
        raise_server_exceptions = self.transport.raise_server_exceptions
        self.transport.raise_server_exceptions = not handle_exceptions

        yield

        self.transport.raise_server_exceptions = raise_server_exceptions



import os
import sys

from argparse import Namespace
from collections.abc import Mapping
from pathlib import Path
from typing import Any
from typing import TextIO

from alembic import util
from alembic.config import Config as AlembicConfig

from expanse.common.core.application import Application


class Config(AlembicConfig):
    def __init__(
        self,
        app: Application,
        file_: str | os.PathLike[str] | None = None,
        ini_section: str = "alembic",
        output_buffer: TextIO | None = None,
        stdout: TextIO = sys.stdout,
        cmd_opts: Namespace | None = None,
        config_args: Mapping[str, Any] = util.immutabledict(),
        attributes: dict[str, Any] | None = None,
    ):
        super().__init__(
            file_=file_
            or app.named_path("database").joinpath("migrations/alembic.ini"),
            ini_section=ini_section,
            output_buffer=output_buffer,
            stdout=stdout,
            cmd_opts=cmd_opts,
            config_args=config_args,
            attributes=attributes,
        )

        self.template_dir = Path(__file__).parent.joinpath("templates")

    def get_template_directory(self) -> str:
        return str(self.template_dir)



from __future__ import annotations

import sys

from typing import TYPE_CHECKING

from expanse.asynchronous.support.service_provider import ServiceProvider
from expanse.common.support._utils import module_from_path


if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path

    from expanse.asynchronous.core.application import Application
    from expanse.asynchronous.routing.router import Router


class RouteServiceProvider(ServiceProvider):
    async def load_routes_from_file(self, router: Router, path: Path) -> Router:
        app: Application = await self._container.get("app")
        module_name = (
            path.resolve()
            .relative_to(app.base_path)
            .with_suffix("")
            .as_posix()
            .replace("/", ".")
        )

        module = module_from_path(path, name=module_name)

        if module is None:
            return router

        # Register route file to sys.modules
        if module_name not in sys.modules:
            sys.modules[module_name] = module

        routes: Callable[[Router], None] = module.routes

        routes(router)

        return router



from pathlib import Path

import pytest

from expanse.common.core.http.exceptions import HTTPException
from expanse.core.application import Application
from expanse.core.helpers import _use_container
from expanse.http.request import Request
from expanse.static.static import Static


FIXTURES_DIR = Path(__file__).parent / "fixtures"


def test_get_returns_file_content_if_file_exists(app: Application) -> None:
    static = Static([FIXTURES_DIR], prefix="/static")

    with app.container.create_scoped_container() as container:
        container.instance(Request, Request.create("http://example.com"))

        with _use_container(container):
            response = static.get("foo.txt")

    assert response.status_code == 200


def test_get_returns_404_not_found_if_file_does_not_exist(app: Application) -> None:
    static = Static([FIXTURES_DIR], prefix="/static")

    with app.container.create_scoped_container() as container:
        container.instance(Request, Request.create("http://example.com"))

        with _use_container(container):
            with pytest.raises(HTTPException) as e:
                static.get("bar.txt")

            assert e.value.status_code == 404



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

import pytest

from expanse.view.view_service_provider import ViewServiceProvider


if TYPE_CHECKING:
    from expanse.core.application import Application


@pytest.fixture(autouse=True)
def setup_app_config(app: Application) -> None:
    config: dict[str, dict[str, Any]] = app.container.get("config")

    config["view"] = {
        "paths": [Path(__file__).parent.joinpath("fixtures/resources/views")]
    }

    app.register(ViewServiceProvider(app.container))



"""
Foo Migration

Revision ID: 1234567891
Revises: 1234567890
Create Date: 2024-09-06 12:34:56
"""

import sqlalchemy as sa

from alembic import op


# revision identifiers, used by Alembic.
revision: str = "1234567891"
down_revision: str | None = "1234567890"


def upgrade() -> None:
    op.add_column(
        "users",
        sa.Column(
            "active", sa.Boolean(), nullable=False, default=True, server_default="true"
        ),
    )


def downgrade() -> None:
    op.drop_column("users", "active")



from collections.abc import Callable
from functools import wraps
from typing import Self

from expanse.container.container import Container
from expanse.http.request import Request
from expanse.http.response import Response
from expanse.types.http.middleware import RequestHandler


class Pipeline:
    """
    A routing pipeline are composed of layers, like middleware, through which
    a request — and corresponding response — must go through.
    """

    def __init__(self, container: Container) -> None:
        self._container = container
        self._pipes: list[Callable[[Request, RequestHandler], Response]] = []
        self._request: Request | None = None

    def use(self, pipes: list[Callable[[Request, RequestHandler], Response]]) -> Self:
        self._pipes = pipes

        return self

    def send(self, request: Request) -> Self:
        self._request = request

        return self

    def to(self, handler: RequestHandler) -> Response:
        from expanse.core.helpers import _set_container

        if not self._request:
            raise ValueError("No request has been set")

        assert self._request is not None

        _set_container(self._container)

        try:
            return self._build_pipeline(handler)(self._request)
        except Exception as e:
            from expanse.contracts.debug.exception_handler import ExceptionHandler

            if not self._container.has(ExceptionHandler):
                raise e

            exception_handler = self._container.get(ExceptionHandler)

            exception_handler.report(e)

            return exception_handler.render(self._request, e)
        finally:
            _set_container(None)

    def _build_pipeline(self, handler: RequestHandler) -> RequestHandler:
        stack = handler

        for pipe in self._pipes[::-1]:
            stack = self._wrap(pipe)(stack)

        return stack

    def _wrap(
        self, pipe: Callable[[Request, RequestHandler], Response]
    ) -> Callable[[RequestHandler], RequestHandler]:
        @wraps(pipe)
        def decorator(next_call: RequestHandler) -> Callable[[Request], Response]:
            @wraps(next_call)
            def handler(request: Request) -> Response:
                return pipe(request, next_call)

            return handler

        return decorator



from __future__ import annotations

from collections.abc import Callable
from typing import Generic
from typing import TypeVar

from cleo.exceptions import CleoCommandNotFoundError
from cleo.exceptions import CleoLogicError

from expanse.common.console.command import Command


TCommand = TypeVar("TCommand", bound=Command)
Factory = Callable[[], TCommand]


class CommandLoader(Generic[TCommand]):
    def __init__(self, factories: dict[str, Factory]) -> None:
        self._factories = factories

    @property
    def names(self) -> list[str]:
        return list(self._factories.keys())

    def has(self, name: str) -> bool:
        return name in self._factories

    def get(self, name: str) -> TCommand:
        if name not in self._factories:
            raise CleoCommandNotFoundError(name)

        factory = self._factories[name]

        return factory()

    def register_factory(
        self, command_name: str, factory: Callable[[], TCommand]
    ) -> None:
        if command_name in self._factories:
            raise CleoLogicError(f'The command "{command_name}" already exists.')

        self._factories[command_name] = factory
